# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import time

from dimos_lcm.sensor_msgs.Imu import Imu as LCMImu

from dimos.msgs.geometry_msgs import Quaternion, Vector3
from dimos.types.timestamped import Timestamped


class Imu(Timestamped):
    """IMU sensor message mirroring ROS sensor_msgs/Imu.

    Contains orientation, angular velocity, and linear acceleration
    with optional covariance matrices (3x3 row-major as flat 9-element lists).
    """

    msg_name = "sensor_msgs.Imu"

    def __init__(
        self,
        angular_velocity: Vector3 | None = None,
        linear_acceleration: Vector3 | None = None,
        orientation: Quaternion | None = None,
        orientation_covariance: list[float] | None = None,
        angular_velocity_covariance: list[float] | None = None,
        linear_acceleration_covariance: list[float] | None = None,
        frame_id: str = "imu_link",
        ts: float | None = None,
    ) -> None:
        self.ts = ts if ts is not None else time.time()  # type: ignore[assignment]
        self.frame_id = frame_id
        self.angular_velocity = angular_velocity or Vector3(0.0, 0.0, 0.0)
        self.linear_acceleration = linear_acceleration or Vector3(0.0, 0.0, 0.0)
        self.orientation = orientation or Quaternion(0.0, 0.0, 0.0, 1.0)
        self.orientation_covariance = orientation_covariance or [0.0] * 9
        self.angular_velocity_covariance = angular_velocity_covariance or [0.0] * 9
        self.linear_acceleration_covariance = linear_acceleration_covariance or [0.0] * 9

    def lcm_encode(self) -> bytes:
        msg = LCMImu()
        [msg.header.stamp.sec, msg.header.stamp.nsec] = self.ros_timestamp()
        msg.header.frame_id = self.frame_id

        msg.orientation.x = self.orientation.x
        msg.orientation.y = self.orientation.y
        msg.orientation.z = self.orientation.z
        msg.orientation.w = self.orientation.w
        msg.orientation_covariance = self.orientation_covariance

        msg.angular_velocity.x = self.angular_velocity.x
        msg.angular_velocity.y = self.angular_velocity.y
        msg.angular_velocity.z = self.angular_velocity.z
        msg.angular_velocity_covariance = self.angular_velocity_covariance

        msg.linear_acceleration.x = self.linear_acceleration.x
        msg.linear_acceleration.y = self.linear_acceleration.y
        msg.linear_acceleration.z = self.linear_acceleration.z
        msg.linear_acceleration_covariance = self.linear_acceleration_covariance

        return msg.lcm_encode()  # type: ignore[no-any-return]

    @classmethod
    def lcm_decode(cls, data: bytes) -> Imu:
        msg = LCMImu.lcm_decode(data)
        ts = msg.header.stamp.sec + (msg.header.stamp.nsec / 1_000_000_000)
        return cls(
            angular_velocity=Vector3(
                msg.angular_velocity.x,
                msg.angular_velocity.y,
                msg.angular_velocity.z,
            ),
            linear_acceleration=Vector3(
                msg.linear_acceleration.x,
                msg.linear_acceleration.y,
                msg.linear_acceleration.z,
            ),
            orientation=Quaternion(
                msg.orientation.x,
                msg.orientation.y,
                msg.orientation.z,
                msg.orientation.w,
            ),
            orientation_covariance=list(msg.orientation_covariance),
            angular_velocity_covariance=list(msg.angular_velocity_covariance),
            linear_acceleration_covariance=list(msg.linear_acceleration_covariance),
            frame_id=msg.header.frame_id,
            ts=ts,
        )

    def __str__(self) -> str:
        return (
            f"Imu(frame_id='{self.frame_id}', "
            f"gyro=({self.angular_velocity.x:.3f}, {self.angular_velocity.y:.3f}, {self.angular_velocity.z:.3f}), "
            f"accel=({self.linear_acceleration.x:.3f}, {self.linear_acceleration.y:.3f}, {self.linear_acceleration.z:.3f}))"
        )

    def __repr__(self) -> str:
        return (
            f"Imu(ts={self.ts}, frame_id='{self.frame_id}', "
            f"angular_velocity={self.angular_velocity}, "
            f"linear_acceleration={self.linear_acceleration}, "
            f"orientation={self.orientation})"
        )



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quest teleoperation module."""

from dimos.teleop.quest.quest_extensions import (
    ArmTeleopModule,
    TwistTeleopModule,
    VisualizingTeleopModule,
    arm_teleop_module,
    twist_teleop_module,
    visualizing_teleop_module,
)
from dimos.teleop.quest.quest_teleop_module import (
    Hand,
    QuestTeleopConfig,
    QuestTeleopModule,
    QuestTeleopStatus,
    quest_teleop_module,
)
from dimos.teleop.quest.quest_types import (
    Buttons,
    QuestControllerState,
    ThumbstickState,
)

__all__ = [
    "ArmTeleopModule",
    "Buttons",
    "Hand",
    "QuestControllerState",
    "QuestTeleopConfig",
    "QuestTeleopModule",
    "QuestTeleopStatus",
    "ThumbstickState",
    "TwistTeleopModule",
    "VisualizingTeleopModule",
    # Blueprints
    "arm_teleop_module",
    "quest_teleop_module",
    "twist_teleop_module",
    "visualizing_teleop_module",
]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Keyboard teleop blueprint for the Piper arm.

Launches the ControlCoordinator (mock adapter + CartesianIK), the
ManipulationModule (Drake/Meshcat visualization), and a pygame keyboard
teleop UI — all wired together via autoconnect.

Usage:
    dimos run keyboard-teleop-piper
"""

from dimos.control.components import HardwareComponent, HardwareType, make_joints
from dimos.control.coordinator import TaskConfig, control_coordinator
from dimos.core.blueprints import autoconnect
from dimos.core.transport import LCMTransport
from dimos.manipulation.manipulation_module import manipulation_module
from dimos.manipulation.planning.spec import RobotModelConfig
from dimos.msgs.geometry_msgs import PoseStamped, Quaternion, Vector3
from dimos.msgs.sensor_msgs import JointState
from dimos.teleop.keyboard.keyboard_teleop_module import keyboard_teleop_module
from dimos.utils.data import LfsPath, get_data

_PIPER_MODEL_PATH = LfsPath("piper_description/mujoco_model/piper_no_gripper_description.xml")
_PIPER_DATA = get_data("piper_description")

# Piper 6-DOF mock sim + keyboard teleop + Drake visualization
keyboard_teleop_piper = autoconnect(
    keyboard_teleop_module(model_path=_PIPER_MODEL_PATH, ee_joint_id=6),
    control_coordinator(
        tick_rate=100.0,
        publish_joint_state=True,
        joint_state_frame_id="coordinator",
        hardware=[
            HardwareComponent(
                hardware_id="arm",
                hardware_type=HardwareType.MANIPULATOR,
                joints=make_joints("arm", 6),
                adapter_type="mock",
            ),
        ],
        tasks=[
            TaskConfig(
                name="cartesian_ik_arm",
                type="cartesian_ik",
                joint_names=[f"arm_joint{i + 1}" for i in range(6)],
                priority=10,
                model_path=_PIPER_MODEL_PATH,
                ee_joint_id=6,
            ),
        ],
    ),
    manipulation_module(
        robots=[
            RobotModelConfig(
                name="arm",
                urdf_path=_PIPER_DATA / "urdf" / "piper_description.xacro",
                base_pose=PoseStamped(
                    position=Vector3(x=0.0, y=0.0, z=0.0),
                    orientation=Quaternion(0.0, 0.0, 0.0, 1.0),
                ),
                joint_names=["joint1", "joint2", "joint3", "joint4", "joint5", "joint6"],
                end_effector_link="gripper_base",
                base_link="base_link",
                package_paths={
                    "piper_description": _PIPER_DATA,
                    "piper_gazebo": _PIPER_DATA,  # xacro refs $(find piper_gazebo); unused by Drake
                },
                joint_name_mapping={f"arm_joint{i}": f"joint{i}" for i in range(1, 7)},
                auto_convert_meshes=True,
                home_joints=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            ),
        ],
        enable_viz=True,
    ),
).transports(
    {
        ("cartesian_command", PoseStamped): LCMTransport(
            "/coordinator/cartesian_command", PoseStamped
        ),
        ("joint_state", JointState): LCMTransport("/coordinator/joint_state", JointState),
    }
)

__all__ = ["keyboard_teleop_piper"]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import cache
import os
from pathlib import Path
import platform
import subprocess
import sys
import tarfile
import tempfile

from dimos.constants import DIMOS_PROJECT_ROOT
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


def _get_user_data_dir() -> Path:
    """Get platform-specific user data directory."""
    system = platform.system()
    # if virtual env is available, use it to keep venv's from fighting over data
    # a better fix for large files will be added later to minimize storage duplication
    if os.environ.get("VIRTUAL_ENV"):
        venv_data_dir = Path(
            f"{os.environ.get('VIRTUAL_ENV')}/lib/python{sys.version_info.major}.{sys.version_info.minor}/site-packages/dimos/data"
        )
        return venv_data_dir

    if system == "Linux":
        # Use XDG_DATA_HOME if set, otherwise default to ~/.local/share
        xdg_data_home = os.environ.get("XDG_DATA_HOME")
        if xdg_data_home:
            return Path(xdg_data_home) / "dimos"
        return Path.home() / ".local" / "share" / "dimos"
    elif system == "Darwin":  # macOS
        return Path.home() / "Library" / "Application Support" / "dimos"
    else:
        # Fallback for other systems
        return Path.home() / ".dimos"


@cache
def _get_repo_root() -> Path:
    # Check if running from git repo
    if (DIMOS_PROJECT_ROOT / ".git").exists():
        return DIMOS_PROJECT_ROOT

    # Running as installed package - clone repo to data dir
    try:
        data_dir = _get_user_data_dir()
        data_dir.mkdir(parents=True, exist_ok=True)
        # Test if writable
        test_file = data_dir / ".write_test"
        test_file.touch()
        test_file.unlink()
        logger.info(f"Using local user data directory at '{data_dir}'")
    except (OSError, PermissionError):
        # Fall back to temp dir if data dir not writable
        data_dir = Path(tempfile.gettempdir()) / "dimos"
        data_dir.mkdir(parents=True, exist_ok=True)
        logger.info(f"Using tmp data directory at '{data_dir}'")

    repo_dir = data_dir / "repo"

    # Clone if not already cloned
    if not (repo_dir / ".git").exists():
        try:
            env = os.environ.copy()
            env["GIT_LFS_SKIP_SMUDGE"] = "1"
            subprocess.run(
                [
                    "git",
                    "clone",
                    "--depth",
                    "1",
                    "--branch",
                    "main",
                    "https://github.com/dimensionalOS/dimos.git",
                    str(repo_dir),
                ],
                check=True,
                capture_output=True,
                text=True,
                env=env,
            )
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"Failed to clone dimos repository: {e.stderr}\n"
                f"Make sure you can access https://github.com/dimensionalOS/dimos.git"
            )

    return repo_dir


@cache
def get_data_dir(extra_path: str | None = None) -> Path:
    if extra_path:
        return _get_repo_root() / "data" / extra_path
    return _get_repo_root() / "data"


@cache
def _get_lfs_dir() -> Path:
    return get_data_dir() / ".lfs"


def _check_git_lfs_available() -> bool:
    missing = []

    # Check if git is available
    try:
        subprocess.run(["git", "--version"], capture_output=True, check=True, text=True)
    except (subprocess.CalledProcessError, FileNotFoundError):
        missing.append("git")

    # Check if git-lfs is available
    try:
        subprocess.run(["git-lfs", "version"], capture_output=True, check=True, text=True)
    except (subprocess.CalledProcessError, FileNotFoundError):
        missing.append("git-lfs")

    if missing:
        raise RuntimeError(
            f"Missing required tools: {', '.join(missing)}.\n\n"
            "Git LFS installation instructions: https://git-lfs.github.io/"
        )

    return True


def _is_lfs_pointer_file(file_path: Path) -> bool:
    try:
        # LFS pointer files are small (typically < 200 bytes) and start with specific text
        if file_path.stat().st_size > 1024:  # LFS pointers are much smaller
            return False

        with open(file_path, encoding="utf-8") as f:
            first_line = f.readline().strip()
            return first_line.startswith("version https://git-lfs.github.com/spec/")

    except (UnicodeDecodeError, OSError):
        return False


def _lfs_pull(file_path: Path, repo_root: Path) -> None:
    try:
        relative_path = file_path.relative_to(repo_root)

        env = os.environ.copy()
        env["GIT_LFS_FORCE_PROGRESS"] = "1"

        subprocess.run(
            ["git", "lfs", "pull", "--include", str(relative_path)],
            cwd=repo_root,
            check=True,
            env=env,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"Failed to pull LFS file {file_path}: {e}")

    return None


def _decompress_archive(filename: str | Path) -> Path:
    target_dir = get_data_dir()
    filename_path = Path(filename)
    with tarfile.open(filename_path, "r:gz") as tar:
        tar.extractall(target_dir)
    return target_dir / filename_path.name.replace(".tar.gz", "")


def _pull_lfs_archive(filename: str | Path) -> Path:
    # Check Git LFS availability first
    _check_git_lfs_available()

    # Find repository root
    repo_root = _get_repo_root()

    # Construct path to test data file
    file_path = _get_lfs_dir() / (str(filename) + ".tar.gz")

    # Check if file exists
    if not file_path.exists():
        raise FileNotFoundError(
            f"Test file '{filename}' not found at {file_path}. "
            f"Make sure the file is committed to Git LFS in the tests/data directory."
        )

    # If it's an LFS pointer file, ensure LFS is set up and pull the file
    if _is_lfs_pointer_file(file_path):
        _lfs_pull(file_path, repo_root)

        # Verify the file was actually downloaded
        if _is_lfs_pointer_file(file_path):
            raise RuntimeError(
                f"Failed to download LFS file '{filename}'. The file is still a pointer after attempting to pull."
            )

    return file_path


def get_data(name: str | Path) -> Path:
    """
    Get the path to a test data, downloading from LFS if needed.

    This function will:
    1. Check that Git LFS is available
    2. Locate the file in the tests/data directory
    3. Initialize Git LFS if needed
    4. Download the file from LFS if it's a pointer file
    5. Return the Path object to the actual file or dir

    Supports nested paths like "dataset/subdir/file.jpg" - will download and
    decompress "dataset" archive but return the full nested path.

    Args:
        name: Name of the test file or dir, optionally with nested path
              (e.g., "lidar_sample.bin" or "dataset/frames/001.png")

    Returns:
        Path: Path object to the test file or dir

    Raises:
        RuntimeError: If Git LFS is not available or LFS operations fail
        FileNotFoundError: If the test file doesn't exist

    Usage:
        # Simple file/dir
        file_path = get_data("sample.bin")

        # Nested path - downloads "dataset" archive, returns path to nested file
        frame = get_data("dataset/frames/001.png")
    """
    data_dir = get_data_dir()
    file_path = data_dir / name

    # already pulled and decompressed, return it directly
    if file_path.exists():
        return file_path

    # extract archive root (first path component) and nested path
    path_parts = Path(name).parts
    archive_name = path_parts[0]
    nested_path = Path(*path_parts[1:]) if len(path_parts) > 1 else None

    # download and decompress the archive root
    archive_path = _decompress_archive(_pull_lfs_archive(archive_name))

    # return full path including nested components
    if nested_path:
        return archive_path / nested_path
    return archive_path


class LfsPath(type(Path())):  # type: ignore[misc]
    """
    A Path subclass that lazily downloads LFS data when accessed.

    This is useful for both lazy loading and differentiating between LFS paths and regular paths.

    This class wraps pathlib.Path and ensures that get_data() is called
    before any meaningful filesystem operation, making LFS data lazy-loaded.

    Usage:
        path = LfsPath("sample_data")
        # No download yet

        with path.open('rb') as f:  # Downloads now if needed
            data = f.read()

        # Or use any Path operation:
        if path.exists():  # Downloads now if needed
            files = list(path.iterdir())
    """

    def __new__(cls, filename: str | Path) -> "LfsPath":
        # Create instance with a placeholder path to satisfy Path.__new__
        # We use "." as a dummy path that always exists
        instance: LfsPath = super().__new__(cls, ".")  # type: ignore[call-arg]
        # Store the actual filename as an instance attribute
        object.__setattr__(instance, "_lfs_filename", filename)
        object.__setattr__(instance, "_lfs_resolved_cache", None)
        return instance

    def _ensure_downloaded(self) -> Path:
        """Ensure the LFS data is downloaded and return the resolved path."""
        cache: Path | None = object.__getattribute__(self, "_lfs_resolved_cache")
        if cache is None:
            filename = object.__getattribute__(self, "_lfs_filename")
            cache = get_data(filename)
            object.__setattr__(self, "_lfs_resolved_cache", cache)
        return cache

    def __getattribute__(self, name: str) -> object:
        # During Path.__new__(), _lfs_filename hasn't been set yet.
        # Fall through to normal Path behavior until construction is complete.
        try:
            object.__getattribute__(self, "_lfs_filename")
        except AttributeError:
            return object.__getattribute__(self, name)

        # After construction, allow access to our internal attributes directly
        if name in ("_lfs_filename", "_lfs_resolved_cache", "_ensure_downloaded"):
            return object.__getattribute__(self, name)

        # For all other attributes, ensure download first then delegate to resolved path
        resolved = object.__getattribute__(self, "_ensure_downloaded")()
        return getattr(resolved, name)

    def __str__(self) -> str:
        """String representation returns resolved path."""
        return str(self._ensure_downloaded())

    def __fspath__(self) -> str:
        """Return filesystem path, downloading from LFS if needed."""
        return str(self._ensure_downloaded())

    def __truediv__(self, other: object) -> Path:
        """Path division operator - returns resolved path."""
        return self._ensure_downloaded() / other  # type: ignore[operator, return-value]

    def __rtruediv__(self, other: object) -> Path:
        """Reverse path division operator."""
        return other / self._ensure_downloaded()  # type: ignore[operator, return-value]



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Teleop blueprints for testing and deployment."""

from dimos.control.blueprints import (
    coordinator_teleop_dual,
    coordinator_teleop_piper,
    coordinator_teleop_xarm6,
)
from dimos.core.blueprints import autoconnect
from dimos.core.transport import LCMTransport
from dimos.msgs.geometry_msgs import PoseStamped
from dimos.teleop.quest.quest_extensions import arm_teleop_module, visualizing_teleop_module
from dimos.teleop.quest.quest_types import Buttons

# -----------------------------------------------------------------------------
# Quest Teleop Blueprints
# -----------------------------------------------------------------------------

# Arm teleop with press-and-hold engage
arm_teleop = autoconnect(
    arm_teleop_module(),
).transports(
    {
        ("left_controller_output", PoseStamped): LCMTransport("/teleop/left_delta", PoseStamped),
        ("right_controller_output", PoseStamped): LCMTransport("/teleop/right_delta", PoseStamped),
        ("buttons", Buttons): LCMTransport("/teleop/buttons", Buttons),
    }
)

# Arm teleop with Rerun visualization
arm_teleop_visualizing = autoconnect(
    visualizing_teleop_module(),
).transports(
    {
        ("left_controller_output", PoseStamped): LCMTransport("/teleop/left_delta", PoseStamped),
        ("right_controller_output", PoseStamped): LCMTransport("/teleop/right_delta", PoseStamped),
        ("buttons", Buttons): LCMTransport("/teleop/buttons", Buttons),
    }
)


# -----------------------------------------------------------------------------
# Teleop wired to Coordinator (TeleopIK)
# -----------------------------------------------------------------------------

# Single XArm6 teleop: right controller -> xarm6
# Usage: dimos run arm-teleop-xarm6

arm_teleop_xarm6 = autoconnect(
    arm_teleop_module(task_names={"right": "teleop_xarm"}),
    coordinator_teleop_xarm6,
).transports(
    {
        ("right_controller_output", PoseStamped): LCMTransport(
            "/coordinator/cartesian_command", PoseStamped
        ),
        ("buttons", Buttons): LCMTransport("/teleop/buttons", Buttons),
    }
)


# Single Piper teleop: left controller -> piper arm
# Usage: dimos run arm-teleop-piper
arm_teleop_piper = autoconnect(
    arm_teleop_module(task_names={"left": "teleop_piper"}),
    coordinator_teleop_piper,
).transports(
    {
        ("left_controller_output", PoseStamped): LCMTransport(
            "/coordinator/cartesian_command", PoseStamped
        ),
        ("buttons", Buttons): LCMTransport("/teleop/buttons", Buttons),
    }
)


# Dual arm teleop: right -> piper, left -> xarm6 (TeleopIK)
arm_teleop_dual = autoconnect(
    arm_teleop_module(task_names={"right": "teleop_piper", "left": "teleop_xarm"}),
    coordinator_teleop_dual,
).transports(
    {
        ("right_controller_output", PoseStamped): LCMTransport(
            "/coordinator/cartesian_command", PoseStamped
        ),
        ("left_controller_output", PoseStamped): LCMTransport(
            "/coordinator/cartesian_command", PoseStamped
        ),
        ("buttons", Buttons): LCMTransport("/teleop/buttons", Buttons),
    }
)


__all__ = [
    "arm_teleop",
    "arm_teleop_dual",
    "arm_teleop_piper",
    "arm_teleop_visualizing",
    "arm_teleop_xarm6",
]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from functools import lru_cache
import importlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dimos.msgs import DimosMsg


@lru_cache(maxsize=256)
def resolve_msg_type(type_name: str) -> type[DimosMsg] | None:
    """Resolve a message type name to its class.

    Args:
        type_name: Type name in format "module.ClassName" (e.g., "geometry_msgs.Vector3")

    Returns:
        The message class or None if not found.
    """
    try:
        module_name, class_name = type_name.rsplit(".", 1)
    except ValueError:
        return None

    # Try different import paths
    import_paths = [
        f"dimos.msgs.{module_name}",
        f"dimos_lcm.{module_name}",
    ]

    for path in import_paths:
        try:
            module = importlib.import_module(path)
            return getattr(module, class_name)  # type: ignore[no-any-return]
        except (ImportError, AttributeError):
            continue

    return None



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Phone teleoperation module for DimOS."""

from dimos.teleop.phone.phone_extensions import (
    SimplePhoneTeleop,
    simple_phone_teleop_module,
)
from dimos.teleop.phone.phone_teleop_module import (
    PhoneTeleopConfig,
    PhoneTeleopModule,
    phone_teleop_module,
)

__all__ = [
    "PhoneTeleopConfig",
    "PhoneTeleopModule",
    "SimplePhoneTeleop",
    "phone_teleop_module",
    "simple_phone_teleop_module",
]



import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "detection2d": [
            "Detection2D",
            "Detection2DBBox",
            "Detection2DPerson",
            "Detection2DPoint",
            "Filter2D",
            "ImageDetections2D",
        ],
        "detection3d": [
            "Detection3D",
            "Detection3DBBox",
            "Detection3DPC",
            "ImageDetections3DPC",
            "PointCloudFilter",
            "height_filter",
            "radius_outlier",
            "raycast",
            "statistical",
        ],
        "imageDetections": ["ImageDetections"],
        "utils": ["TableStr"],
    },
)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MuJoCo simulation engine implementation."""

from __future__ import annotations

import threading
import time
from typing import TYPE_CHECKING

import mujoco
import mujoco.viewer as viewer  # type: ignore[import-untyped,import-not-found]

from dimos.simulation.engines.base import SimulationEngine
from dimos.simulation.utils.xml_parser import JointMapping, build_joint_mappings
from dimos.utils.logging_config import setup_logger

if TYPE_CHECKING:
    from pathlib import Path

    from dimos.msgs.sensor_msgs import JointState

logger = setup_logger()


class MujocoEngine(SimulationEngine):
    """
    MuJoCo simulation engine.

    - starts MuJoCo simulation engine
    - loads robot/environment into simulation
    - applies control commands
    """

    def __init__(self, config_path: Path, headless: bool) -> None:
        super().__init__(config_path=config_path, headless=headless)

        xml_path = self._resolve_xml_path(config_path)
        self._model = mujoco.MjModel.from_xml_path(str(xml_path))
        self._xml_path = xml_path

        self._data = mujoco.MjData(self._model)
        self._joint_mappings = build_joint_mappings(self._xml_path, self._model)
        self._joint_names = [mapping.name for mapping in self._joint_mappings]
        self._num_joints = len(self._joint_names)
        timestep = float(self._model.opt.timestep)
        self._control_frequency = 1.0 / timestep if timestep > 0.0 else 100.0

        self._connected = False
        self._lock = threading.Lock()
        self._stop_event = threading.Event()
        self._sim_thread: threading.Thread | None = None

        self._joint_positions = [0.0] * self._num_joints
        self._joint_velocities = [0.0] * self._num_joints
        self._joint_efforts = [0.0] * self._num_joints

        self._joint_position_targets = [0.0] * self._num_joints
        self._joint_velocity_targets = [0.0] * self._num_joints
        self._joint_effort_targets = [0.0] * self._num_joints
        self._command_mode = "position"
        for i, mapping in enumerate(self._joint_mappings):
            current_pos = self._current_position(mapping)
            self._joint_position_targets[i] = current_pos
            self._joint_positions[i] = current_pos

    def _resolve_xml_path(self, config_path: Path) -> Path:
        if config_path is None:
            raise ValueError("config_path is required for MuJoCo simulation loading")
        resolved = config_path.expanduser()
        xml_path = resolved / "scene.xml" if resolved.is_dir() else resolved
        if not xml_path.exists():
            raise FileNotFoundError(f"MuJoCo XML not found: {xml_path}")
        return xml_path

    def _current_position(self, mapping: JointMapping) -> float:
        if mapping.joint_id is not None and mapping.qpos_adr is not None:
            return float(self._data.qpos[mapping.qpos_adr])
        if mapping.tendon_qpos_adrs:
            return float(
                sum(self._data.qpos[adr] for adr in mapping.tendon_qpos_adrs)
                / len(mapping.tendon_qpos_adrs)
            )
        if mapping.actuator_id is not None:
            return float(self._data.actuator_length[mapping.actuator_id])
        return 0.0

    def _apply_control(self) -> None:
        with self._lock:
            if self._command_mode == "effort":
                targets = list(self._joint_effort_targets)
            elif self._command_mode == "velocity":
                targets = list(self._joint_velocity_targets)
            elif self._command_mode == "position":
                targets = list(self._joint_position_targets)
            for i, mapping in enumerate(self._joint_mappings):
                if mapping.actuator_id is None:
                    continue
                if i < len(targets):
                    self._data.ctrl[mapping.actuator_id] = targets[i]

    def _update_joint_state(self) -> None:
        with self._lock:
            for i, mapping in enumerate(self._joint_mappings):
                if mapping.joint_id is not None:
                    if mapping.qpos_adr is not None:
                        self._joint_positions[i] = float(self._data.qpos[mapping.qpos_adr])
                    if mapping.dof_adr is not None:
                        self._joint_velocities[i] = float(self._data.qvel[mapping.dof_adr])
                        self._joint_efforts[i] = float(self._data.qfrc_actuator[mapping.dof_adr])
                    continue

                if mapping.tendon_qpos_adrs:
                    pos_sum = sum(self._data.qpos[adr] for adr in mapping.tendon_qpos_adrs)
                    count = len(mapping.tendon_qpos_adrs)
                    self._joint_positions[i] = float(pos_sum / count)
                    if mapping.tendon_dof_adrs:
                        vel_sum = sum(self._data.qvel[adr] for adr in mapping.tendon_dof_adrs)
                        self._joint_velocities[i] = float(vel_sum / len(mapping.tendon_dof_adrs))
                    else:
                        self._joint_velocities[i] = 0.0
                elif mapping.actuator_id is not None:
                    self._joint_positions[i] = float(
                        self._data.actuator_length[mapping.actuator_id]
                    )
                    self._joint_velocities[i] = 0.0

                if mapping.actuator_id is not None:
                    self._joint_efforts[i] = float(self._data.actuator_force[mapping.actuator_id])

    def connect(self) -> bool:
        try:
            logger.info(f"{self.__class__.__name__}: connect()")
            with self._lock:
                self._connected = True
                self._stop_event.clear()

            if self._sim_thread is None or not self._sim_thread.is_alive():
                self._sim_thread = threading.Thread(
                    target=self._sim_loop,
                    name=f"{self.__class__.__name__}Sim",
                    daemon=True,
                )
                self._sim_thread.start()
            return True
        except Exception as e:
            logger.error(f"{self.__class__.__name__}: connect() failed: {e}")
            return False

    def disconnect(self) -> bool:
        try:
            logger.info(f"{self.__class__.__name__}: disconnect()")
            with self._lock:
                self._connected = False
            self._stop_event.set()
            if self._sim_thread and self._sim_thread.is_alive():
                self._sim_thread.join(timeout=2.0)
            self._sim_thread = None
            return True
        except Exception as e:
            logger.error(f"{self.__class__.__name__}: disconnect() failed: {e}")
            return False

    def _sim_loop(self) -> None:
        logger.info(f"{self.__class__.__name__}: sim loop started")
        dt = 1.0 / self._control_frequency

        def _step_once(sync_viewer: bool) -> None:
            loop_start = time.time()
            self._apply_control()
            mujoco.mj_step(self._model, self._data)
            if sync_viewer:
                m_viewer.sync()
            self._update_joint_state()

            elapsed = time.time() - loop_start
            sleep_time = dt - elapsed
            if sleep_time > 0:
                time.sleep(sleep_time)

        if self._headless:
            while not self._stop_event.is_set():
                _step_once(sync_viewer=False)
        else:
            with viewer.launch_passive(
                self._model, self._data, show_left_ui=False, show_right_ui=False
            ) as m_viewer:
                while m_viewer.is_running() and not self._stop_event.is_set():
                    _step_once(sync_viewer=True)

        logger.info(f"{self.__class__.__name__}: sim loop stopped")

    @property
    def connected(self) -> bool:
        with self._lock:
            return self._connected

    @property
    def num_joints(self) -> int:
        return self._num_joints

    @property
    def joint_names(self) -> list[str]:
        return list(self._joint_names)

    @property
    def model(self) -> mujoco.MjModel:
        return self._model

    @property
    def joint_positions(self) -> list[float]:
        with self._lock:
            return list(self._joint_positions)

    @property
    def joint_velocities(self) -> list[float]:
        with self._lock:
            return list(self._joint_velocities)

    @property
    def joint_efforts(self) -> list[float]:
        with self._lock:
            return list(self._joint_efforts)

    @property
    def control_frequency(self) -> float:
        return self._control_frequency

    def read_joint_positions(self) -> list[float]:
        return self.joint_positions

    def read_joint_velocities(self) -> list[float]:
        return self.joint_velocities

    def read_joint_efforts(self) -> list[float]:
        return self.joint_efforts

    def write_joint_command(self, command: JointState) -> None:
        if command.position:
            self._command_mode = "position"
            self._set_position_targets(command.position)
            return
        if command.velocity:
            self._command_mode = "velocity"
            self._set_velocity_targets(command.velocity)
            return
        if command.effort:
            self._command_mode = "effort"
            self._set_effort_targets(command.effort)
            return

    def _set_position_targets(self, positions: list[float]) -> None:
        if len(positions) > self._num_joints:
            raise ValueError(
                f"Position command has {len(positions)} joints, expected at most {self._num_joints}"
            )
        with self._lock:
            for i in range(len(positions)):
                self._joint_position_targets[i] = float(positions[i])

    def _set_velocity_targets(self, velocities: list[float]) -> None:
        if len(velocities) > self._num_joints:
            raise ValueError(
                f"Velocity command has {len(velocities)} joints, expected at most {self._num_joints}"
            )
        with self._lock:
            for i in range(len(velocities)):
                self._joint_velocity_targets[i] = float(velocities[i])

    def _set_effort_targets(self, efforts: list[float]) -> None:
        if len(efforts) > self._num_joints:
            raise ValueError(
                f"Effort command has {len(efforts)} joints, expected at most {self._num_joints}"
            )
        with self._lock:
            for i in range(len(efforts)):
                self._joint_effort_targets[i] = float(efforts[i])

    def hold_current_position(self) -> None:
        with self._lock:
            self._command_mode = "position"
            for i, mapping in enumerate(self._joint_mappings):
                self._joint_position_targets[i] = self._current_position(mapping)


__all__ = [
    "MujocoEngine",
]



# Copyright 2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dimos.agents.agent import Agent
from dimos.core.blueprints import autoconnect
from dimos.hardware.sensors.camera import zed
from dimos.hardware.sensors.camera.module import camera_module
from dimos.hardware.sensors.camera.webcam import Webcam

demo_agent = autoconnect(Agent.blueprint())

demo_agent_camera = autoconnect(
    Agent.blueprint(),
    camera_module(
        hardware=lambda: Webcam(
            camera_index=0,
            fps=15,
            camera_info=zed.CameraInfo.SingleWebcam,
        ),
    ),
)



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bridge utilities for connecting pubsub systems."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, Protocol, TypeVar

from dimos.protocol.service.spec import Service

if TYPE_CHECKING:
    from collections.abc import Callable

    from dimos.protocol.pubsub.spec import AllPubSub, PubSub


TopicT = TypeVar("TopicT")
MsgT = TypeVar("MsgT")
TopicFrom = TypeVar("TopicFrom")
TopicTo = TypeVar("TopicTo")
MsgFrom = TypeVar("MsgFrom")
MsgTo = TypeVar("MsgTo")


class Translator(Protocol[TopicFrom, TopicTo, MsgFrom, MsgTo]):  # type: ignore[misc]
    """Protocol for translating topics and messages between pubsub systems."""

    def topic(self, topic: TopicFrom) -> TopicTo:
        """Translate a topic from source to destination format."""
        ...

    def msg(self, msg: MsgFrom) -> MsgTo:
        """Translate a message from source to destination format."""
        ...


def bridge(
    pubsub1: AllPubSub[TopicFrom, MsgFrom],
    pubsub2: PubSub[TopicTo, MsgTo],
    translator: Translator[TopicFrom, TopicTo, MsgFrom, MsgTo],
    # optionally we can override subscribe_all
    # and only bridge a specific part of the pubsub tree
    topic_from: TopicFrom | None = None,
) -> Callable[[], None]:
    def pass_msg(msg: MsgFrom, topic: TopicFrom) -> None:
        pubsub2.publish(translator.topic(topic), translator.msg(msg))

    # Bridge only specific messages from pubsub1 to pubsub2
    if topic_from:
        return pubsub1.subscribe(topic_from, pass_msg)

    # Bridge all messages from pubsub1 to pubsub2
    return pubsub1.subscribe_all(pass_msg)


@dataclass
class BridgeConfig(Generic[TopicFrom, TopicTo, MsgFrom, MsgTo]):
    """Configuration for a one-way bridge."""

    source: AllPubSub[TopicFrom, MsgFrom]
    destination: PubSub[TopicTo, MsgTo]
    translator: Translator[TopicFrom, TopicTo, MsgFrom, MsgTo]
    subscribe_topic: TopicFrom | None = None


class Bridge(Service[BridgeConfig[TopicFrom, TopicTo, MsgFrom, MsgTo]]):
    """Service that bridges messages from one pubsub to another."""

    _unsubscribe: Callable[[], None] | None = None

    def start(self) -> None:
        super().start()
        self._unsubscribe = bridge(
            self.config.source,
            self.config.destination,
            self.config.translator,
            self.config.subscribe_topic,
        )

    def stop(self) -> None:
        if self._unsubscribe:
            self._unsubscribe()
            self._unsubscribe = None
        super().stop()



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compatibility re-exports for legacy dimos.robot.unitree_webrtc.type.* imports."""

import importlib

__all__ = []


def __getattr__(name: str):  # type: ignore[no-untyped-def]
    module = importlib.import_module("dimos.robot.unitree.type")
    try:
        return getattr(module, name)
    except AttributeError as exc:
        raise AttributeError(f"No {__name__} attribute {name}") from exc


def __dir__() -> list[str]:
    module = importlib.import_module("dimos.robot.unitree.type")
    return [name for name in dir(module) if not name.startswith("_")]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MuJoCo XML parsing helpers for joint/actuator metadata."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
import xml.etree.ElementTree as ET

import mujoco

if TYPE_CHECKING:
    from pathlib import Path


@dataclass(frozen=True)
class JointMapping:
    name: str
    joint_id: int | None
    actuator_id: int | None
    qpos_adr: int | None
    dof_adr: int | None
    tendon_qpos_adrs: tuple[int, ...]
    tendon_dof_adrs: tuple[int, ...]


@dataclass(frozen=True)
class _ActuatorSpec:
    name: str
    joint: str | None
    tendon: str | None


def build_joint_mappings(xml_path: Path, model: mujoco.MjModel) -> list[JointMapping]:
    specs = _parse_actuator_specs(xml_path)
    if specs:
        return _build_joint_mappings_from_specs(specs, model)
    if int(model.nu) > 0:
        return _build_joint_mappings_from_actuators(model)
    return _build_joint_mappings_from_model(model)


def _parse_actuator_specs(xml_path: Path) -> list[_ActuatorSpec]:
    return _collect_actuator_specs(xml_path.resolve(), seen=set())


def _collect_actuator_specs(xml_path: Path, seen: set[Path]) -> list[_ActuatorSpec]:
    if xml_path in seen:
        return []
    seen.add(xml_path)

    root = ET.parse(xml_path).getroot()
    base_dir = xml_path.parent
    specs: list[_ActuatorSpec] = []

    def walk(node: ET.Element) -> None:
        for child in node:
            if child.tag == "include":
                include_file = child.attrib.get("file")
                if include_file:
                    include_path = (base_dir / include_file).resolve()
                    specs.extend(_collect_actuator_specs(include_path, seen))
                continue
            if child.tag == "actuator":
                specs.extend(_parse_actuator_block(child))
                continue
            walk(child)

    walk(root)
    return specs


def _parse_actuator_block(actuator_elem: ET.Element) -> list[_ActuatorSpec]:
    specs: list[_ActuatorSpec] = []
    for child in actuator_elem:
        joint = child.attrib.get("joint")
        tendon = child.attrib.get("tendon")
        if not joint and not tendon:
            continue
        name = child.attrib.get("name") or joint or tendon or "actuator"
        specs.append(_ActuatorSpec(name=name, joint=joint, tendon=tendon))
    return specs


def _build_joint_mappings_from_specs(
    specs: list[_ActuatorSpec],
    model: mujoco.MjModel,
) -> list[JointMapping]:
    mappings: list[JointMapping] = []
    for spec in specs:
        if spec.joint:
            mappings.append(_mapping_for_joint(spec, model))
        elif spec.tendon:
            mappings.append(_mapping_for_tendon(spec, model))
    return mappings


def _mapping_for_joint(spec: _ActuatorSpec, model: mujoco.MjModel) -> JointMapping:
    joint_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_JOINT, spec.joint)
    if joint_id < 0:
        raise ValueError(f"Unknown joint '{spec.joint}' in MuJoCo model")
    actuator_id = _find_actuator_id_for_joint(model, joint_id, spec.name)
    joint_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_JOINT, joint_id) or spec.name
    return JointMapping(
        name=joint_name,
        joint_id=joint_id,
        actuator_id=actuator_id,
        qpos_adr=int(model.jnt_qposadr[joint_id]),
        dof_adr=int(model.jnt_dofadr[joint_id]),
        tendon_qpos_adrs=(),
        tendon_dof_adrs=(),
    )


def _mapping_for_tendon(spec: _ActuatorSpec, model: mujoco.MjModel) -> JointMapping:
    name = spec.name or spec.tendon
    if not name:
        raise ValueError("Tendon actuator is missing a name and tendon reference")
    tendon_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_TENDON, spec.tendon)
    if tendon_id < 0:
        raise ValueError(f"Unknown tendon '{spec.tendon}' in MuJoCo model")
    actuator_id = _find_actuator_id_for_tendon(model, tendon_id, spec.name)
    joint_ids = _tendon_joint_ids(model, tendon_id)
    return JointMapping(
        name=name,
        joint_id=None,
        actuator_id=actuator_id,
        qpos_adr=None,
        dof_adr=None,
        tendon_qpos_adrs=tuple(int(model.jnt_qposadr[joint_id]) for joint_id in joint_ids),
        tendon_dof_adrs=tuple(int(model.jnt_dofadr[joint_id]) for joint_id in joint_ids),
    )


def _find_actuator_id_for_joint(
    model: mujoco.MjModel,
    joint_id: int,
    actuator_name: str | None,
) -> int | None:
    if actuator_name:
        act_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_ACTUATOR, actuator_name)
        if act_id >= 0:
            return int(act_id)
    for act_id in range(int(model.nu)):
        trn_type = int(model.actuator_trntype[act_id])
        if trn_type != int(mujoco.mjtTrn.mjTRN_JOINT):
            continue
        if int(model.actuator_trnid[act_id, 0]) == joint_id:
            return act_id
    return None


def _find_actuator_id_for_tendon(
    model: mujoco.MjModel,
    tendon_id: int,
    actuator_name: str | None,
) -> int | None:
    if actuator_name:
        act_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_ACTUATOR, actuator_name)
        if act_id >= 0:
            return int(act_id)
    for act_id in range(int(model.nu)):
        trn_type = int(model.actuator_trntype[act_id])
        if trn_type != int(mujoco.mjtTrn.mjTRN_TENDON):
            continue
        if int(model.actuator_trnid[act_id, 0]) == tendon_id:
            return act_id
    return None


def _tendon_joint_ids(model: mujoco.MjModel, tendon_id: int) -> tuple[int, ...]:
    adr = int(model.tendon_adr[tendon_id])
    num = int(model.tendon_num[tendon_id])
    joint_ids: list[int] = []
    for wrap_id in range(adr, adr + num):
        wrap_type = int(model.wrap_type[wrap_id])
        if wrap_type == int(mujoco.mjtWrap.mjWRAP_JOINT):
            joint_ids.append(int(model.wrap_objid[wrap_id]))
    return tuple(joint_ids)


def _build_joint_mappings_from_actuators(model: mujoco.MjModel) -> list[JointMapping]:
    mappings: list[JointMapping] = []
    for actuator_id in range(int(model.nu)):
        actuator_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_ACTUATOR, actuator_id)
        name = actuator_name or f"actuator{actuator_id}"
        trn_type = int(model.actuator_trntype[actuator_id])
        if trn_type == int(mujoco.mjtTrn.mjTRN_JOINT):
            joint_id = int(model.actuator_trnid[actuator_id, 0])
            joint_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_JOINT, joint_id)
            mappings.append(
                JointMapping(
                    name=joint_name or name,
                    joint_id=joint_id,
                    actuator_id=actuator_id,
                    qpos_adr=int(model.jnt_qposadr[joint_id]),
                    dof_adr=int(model.jnt_dofadr[joint_id]),
                    tendon_qpos_adrs=(),
                    tendon_dof_adrs=(),
                )
            )
            continue

        if trn_type == int(mujoco.mjtTrn.mjTRN_TENDON):
            tendon_id = int(model.actuator_trnid[actuator_id, 0])
            tendon_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_TENDON, tendon_id)
            if not actuator_name and tendon_name:
                name = tendon_name
            joint_ids = _tendon_joint_ids(model, tendon_id)
            mappings.append(
                JointMapping(
                    name=name,
                    joint_id=None,
                    actuator_id=actuator_id,
                    qpos_adr=None,
                    dof_adr=None,
                    tendon_qpos_adrs=tuple(
                        int(model.jnt_qposadr[joint_id]) for joint_id in joint_ids
                    ),
                    tendon_dof_adrs=tuple(
                        int(model.jnt_dofadr[joint_id]) for joint_id in joint_ids
                    ),
                )
            )
            continue

        mappings.append(
            JointMapping(
                name=name,
                joint_id=None,
                actuator_id=actuator_id,
                qpos_adr=None,
                dof_adr=None,
                tendon_qpos_adrs=(),
                tendon_dof_adrs=(),
            )
        )

    return mappings


def _build_joint_mappings_from_model(model: mujoco.MjModel) -> list[JointMapping]:
    mappings: list[JointMapping] = []
    for joint_id in range(int(model.njnt)):
        jnt_type = int(model.jnt_type[joint_id])
        if jnt_type not in (
            int(mujoco.mjtJoint.mjJNT_HINGE),
            int(mujoco.mjtJoint.mjJNT_SLIDE),
        ):
            continue
        joint_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_JOINT, joint_id)
        name = joint_name or f"joint{joint_id}"
        mappings.append(
            JointMapping(
                name=name,
                joint_id=joint_id,
                actuator_id=None,
                qpos_adr=int(model.jnt_qposadr[joint_id]),
                dof_adr=int(model.jnt_dofadr[joint_id]),
                tendon_qpos_adrs=(),
                tendon_dof_adrs=(),
            )
        )
    return mappings



# Copyright 2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python NativeModule wrapper for the FAST-LIO2 + Livox Mid-360 binary.

Binds Livox SDK2 directly into FAST-LIO-NON-ROS for real-time LiDAR SLAM.
Outputs registered (world-frame) point clouds and odometry with covariance.

Usage::

    from dimos.hardware.sensors.lidar.fastlio2.module import FastLio2
    from dimos.core.blueprints import autoconnect

    autoconnect(
        FastLio2.blueprint(host_ip="192.168.1.5"),
        SomeConsumer.blueprint(),
    ).build().loop()
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from dimos.core import Out  # noqa: TC001
from dimos.core.native_module import NativeModule, NativeModuleConfig
from dimos.hardware.sensors.lidar.livox.ports import (
    SDK_CMD_DATA_PORT,
    SDK_HOST_CMD_DATA_PORT,
    SDK_HOST_IMU_DATA_PORT,
    SDK_HOST_LOG_DATA_PORT,
    SDK_HOST_POINT_DATA_PORT,
    SDK_HOST_PUSH_MSG_PORT,
    SDK_IMU_DATA_PORT,
    SDK_LOG_DATA_PORT,
    SDK_POINT_DATA_PORT,
    SDK_PUSH_MSG_PORT,
)
from dimos.msgs.nav_msgs.Odometry import Odometry  # noqa: TC001
from dimos.msgs.sensor_msgs.PointCloud2 import PointCloud2  # noqa: TC001
from dimos.spec import mapping, perception

_CONFIG_DIR = Path(__file__).parent / "config"


@dataclass(kw_only=True)
class FastLio2Config(NativeModuleConfig):
    """Config for the FAST-LIO2 + Livox Mid-360 native module."""

    cwd: str | None = "cpp"
    executable: str = "result/bin/fastlio2_native"
    build_command: str | None = "nix build .#fastlio2_native"

    # Livox SDK hardware config
    host_ip: str = "192.168.1.5"
    lidar_ip: str = "192.168.1.155"
    frequency: float = 10.0

    # Frame IDs for output messages
    frame_id: str = "map"
    child_frame_id: str = "body"

    # FAST-LIO internal processing rates
    msr_freq: float = 50.0
    main_freq: float = 5000.0

    # Output publish rates (Hz)
    pointcloud_freq: float = 10.0
    odom_freq: float = 30.0

    # Point cloud filtering
    voxel_size: float = 0.1
    sor_mean_k: int = 50
    sor_stddev: float = 1.0

    # Global voxel map (disabled when map_freq <= 0)
    map_freq: float = 0.0
    map_voxel_size: float = 0.1
    map_max_range: float = 100.0

    # FAST-LIO YAML config (relative to config/ dir, or absolute path)
    # C++ binary reads YAML directly via yaml-cpp
    config: str = "mid360.yaml"

    # SDK port configuration (see livox/ports.py for defaults)
    cmd_data_port: int = SDK_CMD_DATA_PORT
    push_msg_port: int = SDK_PUSH_MSG_PORT
    point_data_port: int = SDK_POINT_DATA_PORT
    imu_data_port: int = SDK_IMU_DATA_PORT
    log_data_port: int = SDK_LOG_DATA_PORT
    host_cmd_data_port: int = SDK_HOST_CMD_DATA_PORT
    host_push_msg_port: int = SDK_HOST_PUSH_MSG_PORT
    host_point_data_port: int = SDK_HOST_POINT_DATA_PORT
    host_imu_data_port: int = SDK_HOST_IMU_DATA_PORT
    host_log_data_port: int = SDK_HOST_LOG_DATA_PORT

    # Resolved in __post_init__, passed as --config_path to the binary
    config_path: str | None = None

    # config is not a CLI arg (config_path is)
    cli_exclude: frozenset[str] = frozenset({"config"})

    def __post_init__(self) -> None:
        if self.config_path is None:
            path = Path(self.config)
            if not path.is_absolute():
                path = _CONFIG_DIR / path
            self.config_path = str(path.resolve())


class FastLio2(NativeModule, perception.Lidar, perception.Odometry, mapping.GlobalPointcloud):
    """FAST-LIO2 SLAM module with integrated Livox Mid-360 driver.

    Ports:
        lidar (Out[PointCloud2]): World-frame registered point cloud.
        odometry (Out[Odometry]): Pose with covariance at LiDAR scan rate.
        global_map (Out[PointCloud2]): Global voxel map (optional, enable via map_freq > 0).
    """

    default_config: type[FastLio2Config] = FastLio2Config  # type: ignore[assignment]
    lidar: Out[PointCloud2]
    odometry: Out[Odometry]
    global_map: Out[PointCloud2]


fastlio2_module = FastLio2.blueprint

__all__ = [
    "FastLio2",
    "FastLio2Config",
    "fastlio2_module",
]

# Verify protocol port compliance (mypy will flag missing ports)
if TYPE_CHECKING:
    FastLio2()



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""RRT-Connect and RRT* motion planners implementing PlannerSpec.

These planners are backend-agnostic - they only use WorldSpec methods and can work
with any physics backend (Drake, MuJoCo, PyBullet, etc.).
"""

from __future__ import annotations

from dataclasses import dataclass, field
import time
from typing import TYPE_CHECKING

import numpy as np

from dimos.manipulation.planning.spec import (
    JointPath,
    PlanningResult,
    PlanningStatus,
    WorldRobotID,
    WorldSpec,
)
from dimos.manipulation.planning.utils.path_utils import compute_path_length
from dimos.msgs.sensor_msgs import JointState
from dimos.utils.logging_config import setup_logger

if TYPE_CHECKING:
    from numpy.typing import NDArray

logger = setup_logger()


@dataclass(eq=False)
class TreeNode:
    """Node in RRT tree with optional cost tracking (for RRT*)."""

    config: NDArray[np.float64]
    parent: TreeNode | None = None
    children: list[TreeNode] = field(default_factory=list)
    cost: float = 0.0

    def path_to_root(self) -> list[NDArray[np.float64]]:
        """Get path from this node to root."""
        path = []
        node: TreeNode | None = self
        while node is not None:
            path.append(node.config)
            node = node.parent
        return list(reversed(path))


class RRTConnectPlanner:
    """Bi-directional RRT-Connect planner.

    This planner is backend-agnostic - it only uses WorldSpec methods for
    collision checking and can work with any physics backend.
    """

    def __init__(
        self,
        step_size: float = 0.1,
        connect_step_size: float = 0.05,
        goal_tolerance: float = 0.1,
        collision_step_size: float = 0.02,
    ):
        self._step_size = step_size
        self._connect_step_size = connect_step_size
        self._goal_tolerance = goal_tolerance
        self._collision_step_size = collision_step_size

    def plan_joint_path(
        self,
        world: WorldSpec,
        robot_id: WorldRobotID,
        start: JointState,
        goal: JointState,
        timeout: float = 10.0,
        max_iterations: int = 5000,
    ) -> PlanningResult:
        """Plan collision-free path using bi-directional RRT."""
        start_time = time.time()

        # Extract positions as numpy arrays for internal computation
        q_start = np.array(start.position, dtype=np.float64)
        q_goal = np.array(goal.position, dtype=np.float64)
        joint_names = start.name  # Store for converting back to JointState

        error = self._validate_inputs(world, robot_id, start, goal)
        if error is not None:
            return error

        lower, upper = world.get_joint_limits(robot_id)
        start_tree = [TreeNode(config=q_start.copy())]
        goal_tree = [TreeNode(config=q_goal.copy())]
        trees_swapped = False

        for iteration in range(max_iterations):
            if time.time() - start_time > timeout:
                return _create_failure_result(
                    PlanningStatus.TIMEOUT,
                    f"Timeout after {iteration} iterations",
                    time.time() - start_time,
                    iteration,
                )

            sample = np.random.uniform(lower, upper)
            extended = self._extend_tree(
                world, robot_id, start_tree, sample, self._step_size, joint_names
            )

            if extended is not None:
                connected = self._connect_tree(
                    world,
                    robot_id,
                    goal_tree,
                    extended.config,
                    self._connect_step_size,
                    joint_names,
                )
                if connected is not None:
                    path = self._extract_path(extended, connected, joint_names)
                    if trees_swapped:
                        path = list(reversed(path))
                    path = self._simplify_path(world, robot_id, path)
                    return _create_success_result(path, time.time() - start_time, iteration + 1)

            start_tree, goal_tree = goal_tree, start_tree
            trees_swapped = not trees_swapped

        return _create_failure_result(
            PlanningStatus.NO_SOLUTION,
            f"No path found after {max_iterations} iterations",
            time.time() - start_time,
            max_iterations,
        )

    def get_name(self) -> str:
        """Get planner name."""
        return "RRTConnect"

    def _validate_inputs(
        self,
        world: WorldSpec,
        robot_id: WorldRobotID,
        start: JointState,
        goal: JointState,
    ) -> PlanningResult | None:
        """Validate planning inputs, returns error result or None if valid."""
        # Check world is finalized
        if not world.is_finalized:
            return _create_failure_result(
                PlanningStatus.NO_SOLUTION,
                "World must be finalized before planning",
            )

        # Check robot exists
        if robot_id not in world.get_robot_ids():
            return _create_failure_result(
                PlanningStatus.NO_SOLUTION,
                f"Robot '{robot_id}' not found",
            )

        # Check start validity using context-free method
        if not world.check_config_collision_free(robot_id, start):
            return _create_failure_result(
                PlanningStatus.COLLISION_AT_START,
                "Start configuration is in collision",
            )

        # Check goal validity using context-free method
        if not world.check_config_collision_free(robot_id, goal):
            return _create_failure_result(
                PlanningStatus.COLLISION_AT_GOAL,
                "Goal configuration is in collision",
            )

        # Check limits with small tolerance for driver floating-point drift
        lower, upper = world.get_joint_limits(robot_id)
        q_start = np.array(start.position, dtype=np.float64)
        q_goal = np.array(goal.position, dtype=np.float64)
        limit_eps = 1e-3  # ~0.06 degrees

        if np.any(q_start < lower - limit_eps) or np.any(q_start > upper + limit_eps):
            return _create_failure_result(
                PlanningStatus.INVALID_START,
                "Start configuration is outside joint limits",
            )

        if np.any(q_goal < lower - limit_eps) or np.any(q_goal > upper + limit_eps):
            return _create_failure_result(
                PlanningStatus.INVALID_GOAL,
                "Goal configuration is outside joint limits",
            )

        return None

    def _extend_tree(
        self,
        world: WorldSpec,
        robot_id: WorldRobotID,
        tree: list[TreeNode],
        target: NDArray[np.float64],
        step_size: float,
        joint_names: list[str],
    ) -> TreeNode | None:
        """Extend tree toward target, returns new node if successful."""
        # Find nearest node
        nearest = min(tree, key=lambda n: float(np.linalg.norm(n.config - target)))

        # Compute new config
        diff = target - nearest.config
        dist = float(np.linalg.norm(diff))

        if dist <= step_size:
            new_config = target.copy()
        else:
            new_config = nearest.config + step_size * (diff / dist)

        # Check validity of edge using context-free method
        start_state = JointState(name=joint_names, position=nearest.config.tolist())
        end_state = JointState(name=joint_names, position=new_config.tolist())
        if world.check_edge_collision_free(
            robot_id, start_state, end_state, self._collision_step_size
        ):
            new_node = TreeNode(config=new_config, parent=nearest)
            nearest.children.append(new_node)
            tree.append(new_node)
            return new_node

        return None

    def _connect_tree(
        self,
        world: WorldSpec,
        robot_id: WorldRobotID,
        tree: list[TreeNode],
        target: NDArray[np.float64],
        step_size: float,
        joint_names: list[str],
    ) -> TreeNode | None:
        """Try to connect tree to target, returns connected node if successful."""
        # Keep extending toward target
        while True:
            result = self._extend_tree(world, robot_id, tree, target, step_size, joint_names)

            if result is None:
                return None  # Extension failed

            # Check if reached target
            if float(np.linalg.norm(result.config - target)) < self._goal_tolerance:
                return result

    def _extract_path(
        self,
        start_node: TreeNode,
        goal_node: TreeNode,
        joint_names: list[str],
    ) -> JointPath:
        """Extract path from two connected nodes."""
        # Path from start node to its root (reversed to be root->node)
        start_path = start_node.path_to_root()

        # Path from goal node to its root
        goal_path = goal_node.path_to_root()

        # Combine: start_root -> start_node -> goal_node -> goal_root
        # But we need start -> goal, so reverse the goal path
        full_path_arrays = start_path + list(reversed(goal_path))

        # Convert to list of JointState
        return [JointState(name=joint_names, position=q.tolist()) for q in full_path_arrays]

    def _simplify_path(
        self,
        world: WorldSpec,
        robot_id: WorldRobotID,
        path: JointPath,
        max_iterations: int = 100,
    ) -> JointPath:
        """Simplify path by random shortcutting."""
        if len(path) <= 2:
            return path

        simplified = list(path)

        for _ in range(max_iterations):
            if len(simplified) <= 2:
                break

            # Pick two random indices (at least 2 apart)
            i = np.random.randint(0, len(simplified) - 2)
            j = np.random.randint(i + 2, len(simplified))

            # Check if direct connection is valid using context-free method
            # path elements are already JointState
            if world.check_edge_collision_free(
                robot_id, simplified[i], simplified[j], self._collision_step_size
            ):
                # Remove intermediate waypoints
                simplified = simplified[: i + 1] + simplified[j:]

        return simplified


# ============= Result Helpers =============


def _create_success_result(
    path: JointPath,
    planning_time: float,
    iterations: int,
) -> PlanningResult:
    """Create a successful planning result."""
    return PlanningResult(
        status=PlanningStatus.SUCCESS,
        path=path,
        planning_time=planning_time,
        path_length=compute_path_length(path),
        iterations=iterations,
        message="Path found",
    )


def _create_failure_result(
    status: PlanningStatus,
    message: str,
    planning_time: float = 0.0,
    iterations: int = 0,
) -> PlanningResult:
    """Create a failed planning result."""
    return PlanningResult(
        status=status,
        path=[],
        planning_time=planning_time,
        iterations=iterations,
        message=message,
    )



from dimos.models.embedding.base import Embedding, EmbeddingModel

__all__ = [
    "Embedding",
    "EmbeddingModel",
]

# Optional: CLIP support
try:
    from dimos.models.embedding.clip import CLIPModel

    __all__.append("CLIPModel")
except ImportError:
    pass

# Optional: MobileCLIP support
try:
    from dimos.models.embedding.mobileclip import MobileCLIPModel

    __all__.append("MobileCLIPModel")
except ImportError:
    pass

# Optional: TorchReID support
try:
    from dimos.models.embedding.treid import TorchReIDModel

    __all__.append("TorchReIDModel")
except ImportError:
    pass



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import argparse
from contextlib import suppress
from dataclasses import dataclass, field
import importlib
import json
import os
import signal
import subprocess
import threading
import time
from typing import TYPE_CHECKING, Any

from dimos.core.docker_build import build_image, image_exists
from dimos.core.module import Module, ModuleConfig
from dimos.core.rpc_client import RpcCall
from dimos.protocol.rpc import LCMRPC
from dimos.utils.logging_config import setup_logger
from dimos.visualization.rerun.bridge import RERUN_GRPC_PORT, RERUN_WEB_PORT

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

logger = setup_logger()

DOCKER_RUN_TIMEOUT = 120  #     Timeout for `docker run` command execution
DOCKER_CMD_TIMEOUT = 20  #       Timeout for quick Docker commands (inspect, rm, logs)
DOCKER_STATUS_TIMEOUT = 10  #    Timeout for container status checks
DOCKER_STOP_TIMEOUT = 30  #      Timeout for `docker stop` command (graceful shutdown)
RPC_READY_TIMEOUT = 3.0  #       Timeout for RPC readiness probe during container startup
LOG_TAIL_LINES = 200  #          Number of log lines to include in error messages


@dataclass(kw_only=True)
class DockerModuleConfig(ModuleConfig):
    """
    Configuration for running a DimOS module inside Docker.

    For advanced Docker options not listed here, use docker_extra_args.
    Example: docker_extra_args=["--cap-add=SYS_ADMIN", "--read-only"]
    """

    # Build / image
    docker_image: str
    docker_file: Path | None = None  # Required on host for building, not needed in container
    docker_build_context: Path | None = None
    docker_build_args: dict[str, str] = field(default_factory=dict)

    # Identity
    docker_container_name: str | None = None
    docker_labels: dict[str, str] = field(default_factory=dict)

    # Networking (host mode recommended for LCM multicast)
    docker_network_mode: str = "host"
    docker_network: str | None = None
    docker_ports: list[tuple[int, int, str]] = field(
        default_factory=list
    )  # (host, container, proto)

    # Runtime resources
    docker_gpus: str | None = "all"
    docker_shm_size: str = "2g"
    docker_restart_policy: str = "on-failure:3"

    # Env + volumes + devices
    docker_env_files: list[str] = field(default_factory=list)
    docker_env: dict[str, str] = field(default_factory=dict)
    docker_volumes: list[tuple[str, str, str]] = field(
        default_factory=list
    )  # (host, container, mode)
    docker_devices: list[str] = field(default_factory=list)  # --device args as strings

    # Security
    docker_privileged: bool = False

    # Lifecycle / overrides
    docker_rm: bool = False
    docker_entrypoint: str | None = None
    docker_command: list[str] | None = None
    docker_extra_args: list[str] = field(default_factory=list)

    # Startup readiness
    docker_startup_timeout: float = 120.0
    docker_poll_interval: float = 1.0

    # Advanced
    docker_bin: str = "docker"


def is_docker_module(module_class: type) -> bool:
    """Check if a module class should run in Docker based on its default_config."""
    default_config = getattr(module_class, "default_config", None)
    return default_config is not None and issubclass(default_config, DockerModuleConfig)


# Docker helpers


def _run(cmd: list[str], *, timeout: float | None = None) -> subprocess.CompletedProcess[str]:
    logger.debug(f"exec: {' '.join(cmd)}")
    return subprocess.run(cmd, capture_output=True, text=True, timeout=timeout, check=False)


def _docker_bin(cfg: DockerModuleConfig) -> str:
    """Get docker binary path, defaulting to 'docker' if empty/None."""
    return cfg.docker_bin or "docker"


def _remove_container(cfg: DockerModuleConfig, name: str) -> None:
    _run([_docker_bin(cfg), "rm", "-f", name], timeout=DOCKER_CMD_TIMEOUT)


def _is_container_running(cfg: DockerModuleConfig, name: str) -> bool:
    r = _run(
        [_docker_bin(cfg), "inspect", "-f", "{{.State.Running}}", name],
        timeout=DOCKER_STATUS_TIMEOUT,
    )
    return r.returncode == 0 and r.stdout.strip() == "true"


def _tail_logs(cfg: DockerModuleConfig, name: str, n: int = LOG_TAIL_LINES) -> str:
    r = _run([_docker_bin(cfg), "logs", "--tail", str(n), name], timeout=DOCKER_CMD_TIMEOUT)
    out = (r.stdout or "").rstrip()
    err = (r.stderr or "").rstrip()
    return out + ("\n" + err if err else "")


def _extract_module_config(cfg: DockerModuleConfig) -> dict[str, Any]:
    """Extract JSON-serializable config fields for the container (excludes docker_* fields)."""
    out: dict[str, Any] = {}
    for k, v in cfg.__dict__.items():
        if k.startswith("docker_") or isinstance(v, type) or callable(v):
            continue
        try:
            json.dumps(v)
            out[k] = v
        except (TypeError, ValueError):
            logger.debug(f"Config field '{k}' not JSON-serializable, skipping")
    return out


# Host-side Docker-backed Module handle


class DockerModule:
    """
    Host-side handle for a module running inside Docker.

    Lifecycle:
    - start(): launches container, waits for module ready via RPC
    - stop(): stops container
    - __getattr__: exposes RpcCall for @rpc methods on remote module

    Communication: All RPC happens via LCM multicast (requires --network=host).
    """

    def __init__(self, module_class: type[Module], *args: Any, **kwargs: Any) -> None:
        # Config
        config_class = getattr(module_class, "default_config", DockerModuleConfig)
        config = config_class(**kwargs)

        # Module info
        self._module_class = module_class
        self._config = config
        self._args = args
        self._kwargs = kwargs
        self._running = False
        self.remote_name = module_class.__name__
        self._container_name = (
            config.docker_container_name
            or f"dimos_{module_class.__name__.lower()}_{os.getpid()}_{int(time.time())}"
        )

        # RPC setup
        self.rpc = LCMRPC()
        self.rpcs = set(module_class.rpcs.keys())  # type: ignore[attr-defined]
        self.rpc_calls: list[str] = getattr(module_class, "rpc_calls", [])
        self._unsub_fns: list[Callable[[], None]] = []
        self._bound_rpc_calls: dict[str, RpcCall] = {}

        # Build image if needed (but don't start - caller must call start() explicitly)
        if not image_exists(config):
            logger.info(f"Building {config.docker_image}")
            build_image(config)

    def set_rpc_method(self, method: str, callable: RpcCall) -> None:
        callable.set_rpc(self.rpc)
        self._bound_rpc_calls[method] = callable

    def get_rpc_calls(self, *methods: str) -> RpcCall | tuple[RpcCall, ...]:
        # Check all requested methods exist
        missing = set(methods) - self._bound_rpc_calls.keys()
        if missing:
            raise ValueError(f"RPC methods not found: {missing}")
        # Return single RpcCall or tuple
        calls = tuple(self._bound_rpc_calls[m] for m in methods)
        return calls[0] if len(calls) == 1 else calls

    def start(self) -> None:
        if self._running:
            return

        cfg = self._config

        # Prevent accidental kill of running container with same name
        if _is_container_running(cfg, self._container_name):
            raise RuntimeError(
                f"Container '{self._container_name}' already running. "
                "Choose a different container_name or stop the existing container."
            )
        _remove_container(cfg, self._container_name)

        cmd = self._build_docker_run_command()
        logger.info(f"Starting docker container: {self._container_name}")
        r = _run(cmd, timeout=DOCKER_RUN_TIMEOUT)
        if r.returncode != 0:
            raise RuntimeError(
                f"Failed to start container.\nSTDOUT:\n{r.stdout}\nSTDERR:\n{r.stderr}"
            )

        self.rpc.start()
        self._running = True
        self._wait_for_ready()

    def stop(self) -> None:
        """Gracefully stop the Docker container and clean up resources."""
        # Signal remote module, stop RPC, unsubscribe handlers (ignore failures)
        with suppress(Exception):
            if self._running:
                self.rpc.call_nowait(f"{self.remote_name}/stop", ([], {}))
        with suppress(Exception):
            self.rpc.stop()
        for unsub in self._unsub_fns:
            with suppress(Exception):
                unsub()
        self._unsub_fns.clear()

        # Stop and remove container
        _run([_docker_bin(self._config), "stop", self._container_name], timeout=DOCKER_STOP_TIMEOUT)
        _remove_container(self._config, self._container_name)
        self._running = False
        logger.info(f"Stopped container: {self._container_name}")

    def status(self) -> dict[str, Any]:
        cfg = self._config
        return {
            "module": self.remote_name,
            "container_name": self._container_name,
            "image": cfg.docker_image,
            "running": bool(self._running and _is_container_running(cfg, self._container_name)),
        }

    def tail_logs(self, n: int = 200) -> str:
        return _tail_logs(self._config, self._container_name, n=n)

    def set_transport(self, stream_name: str, transport: Any) -> bool:
        """Configure stream transport in container. Mirrors DaskModule.set_transport() for autoconnect()."""
        topic = getattr(transport, "topic", None)
        if topic is None:
            return False
        if hasattr(topic, "topic"):
            topic = topic.topic
        result, _ = self.rpc.call_sync(
            f"{self.remote_name}/configure_stream", ([stream_name, str(topic)], {})
        )
        return bool(result)

    def __getattr__(self, name: str) -> Any:
        if name in self.rpcs:
            original_method = getattr(self._module_class, name, None)
            return RpcCall(original_method, self.rpc, name, self.remote_name, self._unsub_fns, None)
        raise AttributeError(f"{name} not found on {self._module_class.__name__}")

    # Docker command building (split into focused helpers for readability)

    def _build_docker_run_command(self) -> list[str]:
        """Build the complete `docker run` command."""
        cfg = self._config
        self._validate_config(cfg)

        cmd = [_docker_bin(cfg), "run", "-d"]
        self._add_lifecycle_args(cmd, cfg)
        self._add_network_args(cmd, cfg)
        self._add_port_args(cmd, cfg)
        self._add_resource_args(cmd, cfg)
        self._add_security_args(cmd, cfg)
        self._add_device_args(cmd, cfg)
        self._add_label_args(cmd, cfg)
        self._add_env_args(cmd, cfg)
        self._add_volume_args(cmd, cfg)
        self._add_entrypoint_args(cmd, cfg)
        cmd.extend(cfg.docker_extra_args)

        cmd.append(cfg.docker_image)
        cmd.extend(self._build_container_command(cfg))
        return cmd

    def _validate_config(self, cfg: DockerModuleConfig) -> None:
        """Validate config before building command."""
        # Warn about network mode - LCM multicast requires host network
        using_host_network = cfg.docker_network is None and cfg.docker_network_mode == "host"
        if not using_host_network:
            logger.warning(
                "DockerModule not using host network. LCM multicast requires --network=host. "
                "RPC communication may not work with bridge/custom networks."
            )

    def _add_lifecycle_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add --rm and --name args."""
        if cfg.docker_rm:
            cmd.append("--rm")
            if cfg.docker_restart_policy and cfg.docker_restart_policy != "no":
                logger.warning(
                    "--rm with docker_restart_policy is unusual; consider docker_restart_policy='no'."
                )
        cmd.extend(["--name", self._container_name])

    def _add_network_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add --network args."""
        if cfg.docker_network and cfg.docker_network_mode != "host":
            logger.warning(
                "Both 'docker_network' and 'docker_network_mode' set; using 'docker_network' and ignoring 'docker_network_mode'."
            )
        if cfg.docker_network:
            cmd.extend(["--network", cfg.docker_network])
        else:
            cmd.append(f"--network={cfg.docker_network_mode}")

    def _add_port_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add -p port args. No-op for host network (ports auto-exposed)."""
        if cfg.docker_network is None and cfg.docker_network_mode == "host":
            return
        # Non-host network: map Rerun ports + any custom ports
        for port in (RERUN_GRPC_PORT, RERUN_WEB_PORT):
            cmd.extend(["-p", f"{port}:{port}/tcp"])
        for host_port, container_port, proto in cfg.docker_ports:
            cmd.extend(["-p", f"{host_port}:{container_port}/{proto or 'tcp'}"])

    def _add_resource_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add --shm-size, --restart, --gpus args."""
        cmd.append(f"--shm-size={cfg.docker_shm_size}")
        if cfg.docker_restart_policy:
            cmd.append(f"--restart={cfg.docker_restart_policy}")
        if cfg.docker_gpus:
            cmd.extend(["--gpus", cfg.docker_gpus])

    def _add_security_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add --privileged if enabled."""
        if cfg.docker_privileged:
            cmd.append("--privileged")

    def _add_device_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add --device args."""
        for dev in cfg.docker_devices:
            cmd.extend(["--device", dev])

    def _add_label_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add --label args with DimOS defaults."""
        labels = dict(cfg.docker_labels)
        labels.setdefault("dimos.kind", "module")
        labels.setdefault("dimos.module", self._module_class.__name__)
        for k, v in labels.items():
            cmd.extend(["--label", f"{k}={v}"])

    def _add_env_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add -e and --env-file args."""
        cmd.extend(["-e", "PYTHONUNBUFFERED=1"])
        for env_file in cfg.docker_env_files:
            cmd.extend(["--env-file", env_file])
        for k, v in cfg.docker_env.items():
            cmd.extend(["-e", f"{k}={v}"])

    def _add_volume_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add -v volume args."""
        for host_path, container_path, mode in cfg.docker_volumes:
            cmd.extend(["-v", f"{host_path}:{container_path}:{mode}"])

    def _add_entrypoint_args(self, cmd: list[str], cfg: DockerModuleConfig) -> None:
        """Add --entrypoint override."""
        if cfg.docker_entrypoint:
            cmd.extend(["--entrypoint", cfg.docker_entrypoint])

    def _build_container_command(self, cfg: DockerModuleConfig) -> list[str]:
        """Build the container command (module runner or custom)."""
        if cfg.docker_command:
            return list(cfg.docker_command)

        module_path = f"{self._module_class.__module__}.{self._module_class.__name__}"
        # Filter out docker-specific kwargs (paths, etc.) - only pass module config
        kwargs = {"config": _extract_module_config(cfg)}
        payload = {"module_path": module_path, "args": list(self._args), "kwargs": kwargs}
        # DimOS base image entrypoint already runs "dimos.core.docker_runner run"
        return ["--payload", json.dumps(payload, separators=(",", ":"))]

    def _wait_for_ready(self) -> None:
        """Poll the module's RPC endpoint until ready, crashed, or timeout."""
        cfg = self._config
        start_time = time.time()

        logger.info(f"Waiting for {self.remote_name} to be ready...")

        while (time.time() - start_time) < cfg.docker_startup_timeout:
            if not _is_container_running(cfg, self._container_name):
                logs = _tail_logs(cfg, self._container_name)
                raise RuntimeError(f"Container died during startup:\n{logs}")

            try:
                self.rpc.call_sync(
                    f"{self.remote_name}/start", ([], {}), rpc_timeout=RPC_READY_TIMEOUT
                )
                elapsed = time.time() - start_time
                logger.info(f"{self.remote_name} ready ({elapsed:.1f}s)")
                return
            except (TimeoutError, ConnectionError, OSError):
                # Module not ready yet - retry after poll interval
                time.sleep(cfg.docker_poll_interval)

        logs = _tail_logs(cfg, self._container_name)
        raise RuntimeError(
            f"Timeout waiting for {self.remote_name} after {cfg.docker_startup_timeout:.1f}s:\n{logs}"
        )


# Container-side runner


class StandaloneModuleRunner:
    """Runs a module inside Docker container. Blocks until SIGTERM/SIGINT."""

    def __init__(self, module_path: str, args: list[Any], kwargs: dict[str, Any]) -> None:
        self._module_path = module_path
        self._args = args
        self._module: Module | None = None
        self._shutdown = threading.Event()

        # Merge config fields into kwargs (Configurable creates config from these)
        if "config" in kwargs:
            config_dict = kwargs.pop("config")
            kwargs = {**config_dict, **kwargs}
        self._kwargs = kwargs

    def start(self) -> None:
        mod_path, class_name = self._module_path.rsplit(".", 1)
        mod = importlib.import_module(mod_path)
        module_class = getattr(mod, class_name)

        self._module = module_class(*self._args, **self._kwargs)
        logger.info(f"[docker runner] module constructed: {class_name}")

    def stop(self) -> None:
        self._shutdown.set()
        if self._module is not None:
            try:
                self._module.stop()
            except Exception as e:
                logger.error(f"[docker runner] error stopping module: {e}")

    def wait(self) -> None:
        self._shutdown.wait()


def _install_signal_handlers(runner: StandaloneModuleRunner) -> None:
    def shutdown(_sig: int, _frame: Any) -> None:
        runner.stop()

    signal.signal(signal.SIGTERM, shutdown)
    signal.signal(signal.SIGINT, shutdown)


def _cli_run(payload_json: str) -> None:
    payload = json.loads(payload_json)
    runner = StandaloneModuleRunner(
        payload["module_path"],
        payload.get("args", []),
        payload.get("kwargs", {}),
    )
    _install_signal_handlers(runner)
    runner.start()
    runner.wait()


def main(argv: list[str] | None = None) -> None:
    parser = argparse.ArgumentParser(prog="dimos.core.docker_runner")
    sub = parser.add_subparsers(dest="cmd", required=True)

    runp = sub.add_parser("run", help="Run a module inside a container")
    runp.add_argument("--payload", required=True, help="JSON payload with module_path and config")

    args = parser.parse_args(argv)

    if args.cmd == "run":
        _cli_run(args.payload)
        return

    raise ValueError(f"Unknown cmd: {args.cmd}")


if __name__ == "__main__":
    main()


__all__ = [
    "DockerModule",
    "DockerModuleConfig",
    "is_docker_module",
]



import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "detectors": ["Detector", "Yolo2DDetector"],
        "module2D": ["Detection2DModule"],
        "module3D": ["Detection3DModule"],
    },
)



# Copyright 2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from dataclasses import dataclass, field
from typing import cast

import reactivex as rx
from reactivex import operators as ops
from reactivex.observable import Observable

from dimos.core import In, rpc
from dimos.core.module import Module, ModuleConfig
from dimos.models.embedding.base import Embedding, EmbeddingModel
from dimos.models.embedding.clip import CLIPModel
from dimos.msgs.geometry_msgs import PoseStamped
from dimos.msgs.nav_msgs import OccupancyGrid
from dimos.msgs.sensor_msgs import Image
from dimos.msgs.sensor_msgs.Image import Image, sharpness_barrier
from dimos.utils.reactive import getter_hot


@dataclass
class Config(ModuleConfig):
    embedding_model: EmbeddingModel = field(default_factory=CLIPModel)


@dataclass
class SpatialEntry:
    image: Image
    pose: PoseStamped


@dataclass
class SpatialEmbedding(SpatialEntry):
    embedding: Embedding


class EmbeddingMemory(Module[Config]):
    default_config = Config
    config: Config
    color_image: In[Image]
    global_costmap: In[OccupancyGrid]

    _costmap_getter: Callable[[], OccupancyGrid] | None = None

    def get_costmap(self) -> OccupancyGrid:
        if self._costmap_getter is None:
            self._costmap_getter = getter_hot(self.global_costmap.pure_observable())
            self._disposables.add(self._costmap_getter)
        return self._costmap_getter()

    @rpc
    def query_costmap(self, text: str) -> OccupancyGrid:
        costmap = self.get_costmap()
        # overlay costmap with embedding heat
        return costmap

    @rpc
    def start(self) -> None:
        # would be cool if this sharpness_barrier was somehow self-calibrating
        #
        # we need a Governor system, sharpness_barrier frequency shouldn't
        # be a fixed float but an observable that adjusts based on downstream load
        #
        # (also voxel size for mapper for example would benefit from this)
        self.color_image.pure_observable().pipe(
            sharpness_barrier(0.5),
            ops.flat_map(self._try_create_spatial_entry),
            ops.map(self._embed_spatial_entry),
            ops.map(self._store_spatial_entry),
        ).subscribe(print)

    def _try_create_spatial_entry(self, img: Image) -> Observable[SpatialEntry]:
        pose = self.tf.get_pose("world", "base_link")
        if not pose:
            return rx.empty()
        return rx.of(SpatialEntry(image=img, pose=pose))

    def _embed_spatial_entry(self, spatial_entry: SpatialEntry) -> SpatialEmbedding:
        embedding = cast("Embedding", self.config.embedding_model.embed(spatial_entry.image))
        return SpatialEmbedding(
            image=spatial_entry.image,
            pose=spatial_entry.pose,
            embedding=embedding,
        )

    def _store_spatial_entry(self, spatial_embedding: SpatialEmbedding) -> SpatialEmbedding:
        return spatial_embedding

    def query_text(self, query: str) -> list[SpatialEmbedding]:
        self.config.embedding_model.embed_text(query)
        results: list[SpatialEmbedding] = []
        return results



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Kinematics Module

Contains IK solver implementations that use WorldSpec.

## Implementations

- JacobianIK: Backend-agnostic iterative/differential IK (works with any WorldSpec)
- DrakeOptimizationIK: Drake-specific nonlinear optimization IK (requires DrakeWorld)

## Usage

Use factory functions to create IK solvers:

```python
from dimos.manipulation.planning.factory import create_kinematics

# Backend-agnostic (works with any WorldSpec)
kinematics = create_kinematics(name="jacobian")

# Drake-specific (requires DrakeWorld, more accurate)
kinematics = create_kinematics(name="drake_optimization")

result = kinematics.solve(world, robot_id, target_pose)
```
"""

from dimos.manipulation.planning.kinematics.drake_optimization_ik import (
    DrakeOptimizationIK,
)
from dimos.manipulation.planning.kinematics.jacobian_ik import JacobianIK
from dimos.manipulation.planning.kinematics.pinocchio_ik import (
    PinocchioIK,
    PinocchioIKConfig,
)

__all__ = ["DrakeOptimizationIK", "JacobianIK", "PinocchioIK", "PinocchioIKConfig"]



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quest teleop module extensions and subclasses.

Available subclasses:
    - ArmTeleopModule: Per-hand press-and-hold engage (X/A hold to track), task name routing
    - TwistTeleopModule: Outputs Twist instead of PoseStamped
    - VisualizingTeleopModule: Adds Rerun visualization (inherits press-and-hold engage)
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from dimos.msgs.geometry_msgs import PoseStamped, TwistStamped
from dimos.teleop.quest.quest_teleop_module import Hand, QuestTeleopConfig, QuestTeleopModule
from dimos.teleop.utils.teleop_visualization import (
    visualize_buttons,
    visualize_pose,
)

if TYPE_CHECKING:
    from dimos.core import Out


@dataclass
class TwistTeleopConfig(QuestTeleopConfig):
    """Configuration for TwistTeleopModule."""

    linear_scale: float = 1.0
    angular_scale: float = 1.0


# Example implementation to show how to extend QuestTeleopModule for different teleop behaviors and outputs.
class TwistTeleopModule(QuestTeleopModule):
    """Quest teleop that outputs TwistStamped instead of PoseStamped.

    Config:
        - linear_scale: Scale factor for linear (position) values. Default 1.0.
        - angular_scale: Scale factor for angular (orientation) values. Default 1.0.

    Outputs:
        - left_twist: TwistStamped (linear + angular velocity)
        - right_twist: TwistStamped (linear + angular velocity)
        - buttons: Buttons (inherited)
    """

    default_config = TwistTeleopConfig
    config: TwistTeleopConfig

    left_twist: Out[TwistStamped]
    right_twist: Out[TwistStamped]

    def _publish_msg(self, hand: Hand, output_msg: PoseStamped) -> None:
        """Convert PoseStamped to TwistStamped, apply scaling, and publish."""
        twist = TwistStamped(
            ts=output_msg.ts,
            frame_id=output_msg.frame_id,
            linear=output_msg.position * self.config.linear_scale,
            angular=output_msg.orientation.to_euler() * self.config.angular_scale,
        )
        if hand == Hand.LEFT:
            self.left_twist.publish(twist)
        else:
            self.right_twist.publish(twist)


@dataclass
class ArmTeleopConfig(QuestTeleopConfig):
    """Configuration for ArmTeleopModule.

    Attributes:
        task_names: Mapping of Hand -> coordinator task name. Used to set
            frame_id on output PoseStamped so the coordinator routes each
            hand's commands to the correct TeleopIKTask.
    """

    task_names: dict[str, str] = field(default_factory=dict)


class ArmTeleopModule(QuestTeleopModule):
    """Quest teleop with per-hand press-and-hold engage and task name routing.

    Each controller's primary button (X for left, A for right)
    engages that hand while held, disengages on release.

    When task_names is configured, output PoseStamped messages have their
    frame_id set to the task name, enabling the coordinator to route
    each hand's commands to the correct TeleopIKTask.

    Outputs:
        - left_controller_output: PoseStamped (inherited)
        - right_controller_output: PoseStamped (inherited)
        - buttons: Buttons (inherited)
    """

    default_config = ArmTeleopConfig
    config: ArmTeleopConfig

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._task_names: dict[Hand, str] = {
            Hand[k.upper()]: v for k, v in self.config.task_names.items()
        }

    def _publish_msg(self, hand: Hand, output_msg: PoseStamped) -> None:
        """Stamp frame_id with task name and publish."""
        task_name = self._task_names.get(hand)
        if task_name:
            output_msg = PoseStamped(
                position=output_msg.position,
                orientation=output_msg.orientation,
                ts=output_msg.ts,
                frame_id=task_name,
            )
        super()._publish_msg(hand, output_msg)


class VisualizingTeleopModule(ArmTeleopModule):
    """Quest teleop with Rerun visualization.

    Adds visualization of controller poses and trigger values to Rerun.
    Useful for debugging and development.

    Outputs:
        - left_controller_output: PoseStamped (inherited)
        - right_controller_output: PoseStamped (inherited)
        - buttons: Buttons (inherited)
    """

    def _get_output_pose(self, hand: Hand) -> PoseStamped | None:
        """Get output pose and visualize in Rerun."""
        output_pose = super()._get_output_pose(hand)

        if output_pose is not None:
            current_pose = self._current_poses.get(hand)
            controller = self._controllers.get(hand)
            if current_pose is not None:
                label = "left" if hand == Hand.LEFT else "right"
                visualize_pose(current_pose, label)

                if controller:
                    visualize_buttons(
                        label,
                        primary=controller.primary,
                        secondary=controller.secondary,
                        grip=controller.grip,
                        trigger=controller.trigger,
                    )
        return output_pose


# Module blueprints for easy instantiation
twist_teleop_module = TwistTeleopModule.blueprint
arm_teleop_module = ArmTeleopModule.blueprint
visualizing_teleop_module = VisualizingTeleopModule.blueprint

__all__ = [
    "ArmTeleopConfig",
    "ArmTeleopModule",
    "TwistTeleopModule",
    "VisualizingTeleopModule",
    "arm_teleop_module",
    "twist_teleop_module",
    "visualizing_teleop_module",
]



from .utils import (
    BoundingBox2D,
    CameraInfo,
    Detection2D,
    Detection3D,
    Header,
    Image,
    ObjectData,
    Pose,
    Quaternion,
    Union,
    Vector,
    Vector3,
    bbox2d_to_corners,
    colorize_depth,
    combine_object_data,
    cp,
    cv2,
    detection_results_to_object_data,
    draw_bounding_box,
    draw_object_detection_visualization,
    draw_segmentation_mask,
    extract_pose_from_detection3d,
    find_clicked_detection,
    load_camera_info,
    load_camera_info_opencv,
    logger,
    np,
    point_in_bbox,
    project_2d_points_to_3d,
    project_2d_points_to_3d_cpu,
    project_2d_points_to_3d_cuda,
    project_3d_points_to_2d,
    project_3d_points_to_2d_cpu,
    project_3d_points_to_2d_cuda,
    rectify_image,
    setup_logger,
    torch,
    yaml,
)

__all__ = [
    "BoundingBox2D",
    "CameraInfo",
    "Detection2D",
    "Detection3D",
    "Header",
    "Image",
    "ObjectData",
    "Pose",
    "Quaternion",
    "Union",
    "Vector",
    "Vector3",
    "bbox2d_to_corners",
    "colorize_depth",
    "combine_object_data",
    "cp",
    "cv2",
    "detection_results_to_object_data",
    "draw_bounding_box",
    "draw_object_detection_visualization",
    "draw_segmentation_mask",
    "extract_pose_from_detection3d",
    "find_clicked_detection",
    "load_camera_info",
    "load_camera_info_opencv",
    "logger",
    "np",
    "point_in_bbox",
    "project_2d_points_to_3d",
    "project_2d_points_to_3d_cpu",
    "project_2d_points_to_3d_cuda",
    "project_3d_points_to_2d",
    "project_3d_points_to_2d_cpu",
    "project_3d_points_to_2d_cuda",
    "rectify_image",
    "setup_logger",
    "torch",
    "yaml",
]



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Phone Teleoperation Module.

Receives raw sensor data (TwistStamped) and button state (Bool) from the
phone web app via the Deno LCM bridge.  Computes orientation deltas from
a initial orientation captured on engage, converts to TwistStamped velocity
commands via configurable gains, and publishes.

"""

from dataclasses import dataclass
from pathlib import Path
import shutil
import signal
import subprocess
import threading
import time
from typing import Any

from reactivex.disposable import Disposable

from dimos.core import In, Module, Out, rpc
from dimos.core.module import ModuleConfig
from dimos.msgs.geometry_msgs import Twist, TwistStamped, Vector3
from dimos.msgs.std_msgs.Bool import Bool
from dimos.utils.logging_config import setup_logger

logger = setup_logger()


@dataclass
class PhoneTeleopConfig(ModuleConfig):
    control_loop_hz: float = 50.0
    linear_gain: float = 1.0 / 30.0  # Gain: maps degrees of tilt to m/s. 30 deg -> 1.0 m/s
    angular_gain: float = 1.0 / 30.0  # Gain: maps gyro deg/s to rad/s. 30 deg/s -> 1.0 rad/s


class PhoneTeleopModule(Module[PhoneTeleopConfig]):
    """
    Receives raw sensor data from the phone web app:
      - TwistStamped: linear=(roll, pitch, yaw) deg, angular=(gyro) deg/s
      - Bool: teleop button state (True = held)

    Outputs:
        - twist_output: TwistStamped (velocity command for robot)
    """

    default_config = PhoneTeleopConfig

    # Inputs from Deno bridge
    phone_sensors: In[TwistStamped]
    phone_button: In[Bool]
    # Output: velocity command to robot
    twist_output: Out[TwistStamped]

    # -------------------------------------------------------------------------
    # Initialization
    # -------------------------------------------------------------------------

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self._is_engaged: bool = False
        self._teleop_button: bool = False
        self._current_sensors: TwistStamped | None = None
        self._initial_sensors: TwistStamped | None = None
        self._lock = threading.RLock()

        # Control loop
        self._control_loop_thread: threading.Thread | None = None
        self._stop_event = threading.Event()

        # Deno bridge server
        self._server_process: subprocess.Popen[bytes] | None = None
        self._server_script = Path(__file__).parent / "web" / "teleop_server.ts"

    # -------------------------------------------------------------------------
    # Lifecycle
    # -------------------------------------------------------------------------

    @rpc
    def start(self) -> None:
        super().start()
        for stream, handler in (
            (self.phone_sensors, self._on_sensors),
            (self.phone_button, self._on_button),
        ):
            self._disposables.add(Disposable(stream.subscribe(handler)))  # type: ignore[attr-defined]
        self._start_server()
        self._start_control_loop()

    @rpc
    def stop(self) -> None:
        self._stop_control_loop()
        self._stop_server()
        super().stop()

    # -------------------------------------------------------------------------
    # Internal engage / disengage (assumes lock is held)
    # -------------------------------------------------------------------------

    def _engage(self) -> bool:
        """Engage: capture current sensors as initial"""
        if self._current_sensors is None:
            logger.error("Engage failed: no sensor data yet")
            return False
        self._initial_sensors = self._current_sensors
        self._is_engaged = True
        logger.info("Phone teleop engaged")
        return True

    def _disengage(self) -> None:
        """Disengage: stop publishing"""
        self._is_engaged = False
        self._initial_sensors = None
        logger.info("Phone teleop disengaged")

    # -------------------------------------------------------------------------
    # Callbacks
    # -------------------------------------------------------------------------

    def _on_sensors(self, msg: TwistStamped) -> None:
        """Callback for raw sensor TwistStamped from the phone"""
        with self._lock:
            self._current_sensors = msg

    def _on_button(self, msg: Bool) -> None:
        """Callback for teleop button state."""
        with self._lock:
            self._teleop_button = bool(msg.data)

    # -------------------------------------------------------------------------
    # Deno Bridge Server
    # -------------------------------------------------------------------------

    def _start_server(self) -> None:
        """Launch the Deno WebSocket-to-LCM bridge server as a subprocess."""
        if self._server_process is not None and self._server_process.poll() is None:
            logger.warning("Deno bridge already running", pid=self._server_process.pid)
            return

        if shutil.which("deno") is None:
            logger.error(
                "Deno is not installed. Install it with: curl -fsSL https://deno.land/install.sh | sh"
            )
            return

        script = str(self._server_script)
        cmd = [
            "deno",
            "run",
            "--allow-net",
            "--allow-read",
            "--allow-run",
            "--allow-write",
            "--unstable-net",
            script,
        ]
        try:
            self._server_process = subprocess.Popen(cmd)
            logger.info(f"Deno bridge server started (pid {self._server_process.pid})")
        except OSError as e:
            logger.error(f"Failed to start Deno bridge: {e}")

    def _stop_server(self) -> None:
        """Terminate the Deno bridge server subprocess."""
        if self._server_process is None or self._server_process.poll() is not None:
            self._server_process = None
            return

        logger.info("Stopping Deno bridge server", pid=self._server_process.pid)
        self._server_process.send_signal(signal.SIGTERM)
        try:
            self._server_process.wait(timeout=3)
        except subprocess.TimeoutExpired:
            logger.warning(
                "Deno bridge did not exit, sending SIGKILL", pid=self._server_process.pid
            )
            self._server_process.kill()
            try:
                self._server_process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                logger.error("Deno bridge did not exit after SIGKILL")
        logger.info("Deno bridge server stopped")
        self._server_process = None

    # -------------------------------------------------------------------------
    # Control Loop
    # -------------------------------------------------------------------------

    def _start_control_loop(self) -> None:
        if self._control_loop_thread is not None and self._control_loop_thread.is_alive():
            return

        self._stop_event.clear()
        self._control_loop_thread = threading.Thread(
            target=self._control_loop,
            daemon=True,
            name="PhoneTeleopControlLoop",
        )
        self._control_loop_thread.start()
        logger.info(f"Control loop started at {self.config.control_loop_hz} Hz")

    def _stop_control_loop(self) -> None:
        self._stop_event.set()
        if self._control_loop_thread is not None:
            self._control_loop_thread.join(timeout=1.0)
            self._control_loop_thread = None
        logger.info("Control loop stopped")

    def _control_loop(self) -> None:
        period = 1.0 / self.config.control_loop_hz

        while not self._stop_event.is_set():
            loop_start = time.perf_counter()
            with self._lock:
                self._handle_engage()

                if self._is_engaged:
                    output_twist = self._get_output_twist()
                    if output_twist is not None:
                        self._publish_msg(output_twist)

            elapsed = time.perf_counter() - loop_start
            sleep_time = period - elapsed
            if sleep_time > 0:
                self._stop_event.wait(sleep_time)

    # -------------------------------------------------------------------------
    # Control Loop Internal Methods
    # -------------------------------------------------------------------------

    def _handle_engage(self) -> None:
        """
        Override to customize engagement logic.
        Default: button hold = engaged, release = disengaged.
        """
        if self._teleop_button:
            if not self._is_engaged:
                self._engage()
        else:
            if self._is_engaged:
                self._disengage()

    def _get_output_twist(self) -> TwistStamped | None:
        """Compute twist from orientation delta.
        Override to customize twist computation (e.g., apply scaling, filtering).
        Default: Computes delta angles from initial orientation, applies gains.
        """
        current = self._current_sensors
        initial = self._initial_sensors
        if current is None or initial is None:
            return None

        delta: Twist = Twist(current) - Twist(initial)

        # Handle yaw wraparound (linear.z = yaw, 0-360 degrees)
        d_yaw = delta.linear.z
        if d_yaw > 180:
            d_yaw -= 360
        elif d_yaw < -180:
            d_yaw += 360

        cfg = self.config
        return TwistStamped(
            ts=current.ts,
            frame_id="phone",
            linear=Vector3(
                x=-delta.linear.y * cfg.linear_gain,  # pitch forward -> drive forward
                y=-delta.linear.x * cfg.linear_gain,  # roll right -> strafe right
                z=d_yaw * cfg.linear_gain,  # yaw delta
            ),
            angular=Vector3(
                x=current.angular.x * cfg.angular_gain,
                y=current.angular.y * cfg.angular_gain,
                z=current.angular.z * cfg.angular_gain,
            ),
        )

    def _publish_msg(self, output_msg: TwistStamped) -> None:
        """
        Override to customize output (e.g., apply limits, remap axes).
        """
        self.twist_output.publish(output_msg)


phone_teleop_module = PhoneTeleopModule.blueprint

__all__ = [
    "PhoneTeleopConfig",
    "PhoneTeleopModule",
    "phone_teleop_module",
]



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import asyncio
from collections.abc import AsyncIterator, Callable
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import Any, Generic, Protocol, TypeVar, runtime_checkable

MsgT = TypeVar("MsgT")
TopicT = TypeVar("TopicT")
MsgT_co = TypeVar("MsgT_co", covariant=True)
TopicT_co = TypeVar("TopicT_co", covariant=True)


class PubSubBaseMixin(Generic[TopicT, MsgT]):
    """Mixin class providing sugar methods for PubSub implementations.
    Depends on the basic publish and subscribe methods being implemented.
    """

    def subscribe(
        self, topic: TopicT, callback: Callable[[MsgT, TopicT], None]
    ) -> Callable[[], None]:
        """Subscribe to a topic. Implemented by subclasses."""
        raise NotImplementedError

    @dataclass(slots=True)
    class _Subscription:
        _bus: "PubSubBaseMixin[Any, Any]"
        _topic: Any
        _cb: Callable[[Any, Any], None]
        _unsubscribe_fn: Callable[[], None]

        def unsubscribe(self) -> None:
            self._unsubscribe_fn()

        def __enter__(self) -> "PubSubBaseMixin._Subscription":
            return self

        def __exit__(self, *exc: Any) -> None:
            self.unsubscribe()

    def sub(self, topic: TopicT, cb: Callable[[MsgT, TopicT], None]) -> "_Subscription":
        unsubscribe_fn = self.subscribe(topic, cb)
        return self._Subscription(self, topic, cb, unsubscribe_fn)

    async def aiter(self, topic: TopicT, *, max_pending: int | None = None) -> AsyncIterator[MsgT]:
        q: asyncio.Queue[MsgT] = asyncio.Queue(maxsize=max_pending or 0)

        def _cb(msg: MsgT, topic: TopicT) -> None:
            q.put_nowait(msg)

        unsubscribe_fn = self.subscribe(topic, _cb)
        try:
            while True:
                yield await q.get()
        finally:
            unsubscribe_fn()

    @asynccontextmanager
    async def queue(
        self, topic: TopicT, *, max_pending: int | None = None
    ) -> AsyncIterator[asyncio.Queue[MsgT]]:
        q: asyncio.Queue[MsgT] = asyncio.Queue(maxsize=max_pending or 0)

        def _queue_cb(msg: MsgT, topic: TopicT) -> None:
            q.put_nowait(msg)

        unsubscribe_fn = self.subscribe(topic, _queue_cb)
        try:
            yield q
        finally:
            unsubscribe_fn()


class PubSub(PubSubBaseMixin[TopicT, MsgT], ABC):
    """Abstract base class for pub/sub implementations with sugar methods."""

    @abstractmethod
    def publish(self, topic: TopicT, message: MsgT) -> None:
        """Publish a message to a topic."""
        ...

    @abstractmethod
    def subscribe(
        self, topic: TopicT, callback: Callable[[MsgT, TopicT], None]
    ) -> Callable[[], None]:
        """Subscribe to a topic with a callback. returns unsubscribe function"""
        ...


# AllPubSub and DiscoveryPubSub are complementary mixins:
#
# AllPubsub supports subscribing to all topics (Redis, LCM, MQTT)
# DiscoveryPubSub supports discovering new topics (ROS)
#
# These capabilities are orthogonal but they can implement one another.
# Implementations should subclass whichever matches their native capability.
# The other method will be synthesized automatically.
#
# - AllPubSub: Native support for subscribing to all topics at once.
#   Provides a default subscribe_new_topics() by tracking seen topics.
#
# - DiscoveryPubSub: Native support for discovering new topics as they appear.
#   Provides a default subscribe_all() by subscribing to each discovered topic.
class AllPubSub(PubSub[TopicT, MsgT], ABC):
    """Mixin for PubSub that supports subscribing to all topics.

    Subclass from this if you support native subscribe-all (e.g. MQTT #, Redis *).
    Provides a default subscribe_new_topics() implementation.
    """

    @abstractmethod
    def subscribe_all(self, callback: Callable[[MsgT, TopicT], Any]) -> Callable[[], None]:
        """Subscribe to all topics."""
        ...

    def subscribe_new_topics(self, callback: Callable[[TopicT], Any]) -> Callable[[], None]:
        """Discover new topics by tracking seen topics from subscribe_all."""
        import threading

        seen: set[TopicT] = set()
        lock = threading.Lock()

        def on_msg(msg: MsgT, topic: TopicT) -> None:
            with lock:
                if topic not in seen:
                    seen.add(topic)
                    callback(topic)

        return self.subscribe_all(on_msg)


# This is for ros for now
class DiscoveryPubSub(PubSub[TopicT, MsgT], ABC):
    """Mixin for PubSub that supports discovery of topics.

    Subclass from this if you support topic discovery (e.g. MQTT, Redis, NATS, RabbitMQ).
    """

    @abstractmethod
    def subscribe_new_topics(self, callback: Callable[[TopicT], Any]) -> Callable[[], None]:
        """Get notified when new topics are discovered."""
        ...

    def subscribe_all(self, callback: Callable[[MsgT, TopicT], Any]) -> Callable[[], None]:
        """Subscribe to all topics by subscribing to each discovered topic."""
        import threading

        subscriptions: list[Callable[[], None]] = []
        lock = threading.Lock()

        def on_new_topic(topic: TopicT) -> None:
            unsub = self.subscribe(topic, callback)
            with lock:
                subscriptions.append(unsub)

        discovery_unsub = self.subscribe_new_topics(on_new_topic)

        def unsubscribe_all() -> None:
            discovery_unsub()
            with lock:
                subs = subscriptions.copy()
            for unsub in subs:
                unsub()

        return unsubscribe_all


@runtime_checkable
class SubscribeAllCapable(Protocol[MsgT_co, TopicT_co]):
    """Protocol for pubsubs that support subscribe_all.

    Both AllPubSub (native) and DiscoveryPubSub (synthesized) satisfy this.
    """

    def subscribe_all(self, callback: Callable[[Any, Any], Any]) -> Callable[[], None]:
        """Subscribe to all messages on all topics."""
        ...



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Grid tests for subscribe_all pattern subscriptions."""

from collections.abc import Callable, Generator
from contextlib import AbstractContextManager, contextmanager
from dataclasses import dataclass, field
import re
import time
from typing import Any, Generic, TypeVar

import pytest

from dimos.msgs.geometry_msgs import Pose, Quaternion, Vector3
from dimos.protocol.pubsub.impl.lcmpubsub import LCM, LCMPubSubBase, Topic
from dimos.protocol.pubsub.patterns import Glob
from dimos.protocol.pubsub.spec import AllPubSub, PubSub

TopicT = TypeVar("TopicT")
MsgT = TypeVar("MsgT")

# Type alias for (publisher, subscriber) tuple
PubSubPair = tuple[PubSub[TopicT, MsgT], AllPubSub[TopicT, MsgT]]


@dataclass
class Case(Generic[TopicT, MsgT]):
    """Test case for grid testing pubsub implementations."""

    name: str
    pubsub_context: Callable[[], AbstractContextManager[PubSubPair[TopicT, MsgT]]]
    topic_values: list[tuple[TopicT, MsgT]]
    tags: set[str] = field(default_factory=set)
    # Pattern tests: (pattern_topic, {indices of topic_values that should match})
    glob_patterns: list[tuple[TopicT, set[int]]] = field(default_factory=list)
    regex_patterns: list[tuple[TopicT, set[int]]] = field(default_factory=list)


@contextmanager
def lcm_typed_context() -> Generator[tuple[LCM, LCM], None, None]:
    pub = LCM(autoconf=True)
    sub = LCM(autoconf=False)
    pub.start()
    sub.start()
    try:
        yield pub, sub
    finally:
        pub.stop()
        sub.stop()


@contextmanager
def lcm_bytes_context() -> Generator[tuple[LCMPubSubBase, LCMPubSubBase], None, None]:
    pub = LCMPubSubBase(autoconf=True)
    sub = LCMPubSubBase(autoconf=False)
    pub.start()
    sub.start()
    try:
        yield pub, sub
    finally:
        pub.stop()
        sub.stop()


testcases: list[Case[Any, Any]] = [
    Case(
        name="lcm_typed",
        pubsub_context=lcm_typed_context,
        topic_values=[
            (Topic("/sensor/position", Vector3), Vector3(1, 2, 3)),
            (Topic("/sensor/orientation", Quaternion), Quaternion(0, 0, 0, 1)),
            (Topic("/robot/arm", Pose), Pose(Vector3(4, 5, 6), Quaternion(0, 0, 0, 1))),
        ],
        tags={"all", "glob", "regex"},
        glob_patterns=[
            (Topic(topic=Glob("/sensor/*")), {0, 1}),
            (Topic(topic=Glob("/**/arm")), {2}),
            (Topic(topic=Glob("/**")), {0, 1, 2}),
        ],
        regex_patterns=[
            (Topic(re.compile(r"/sensor/.*")), {0, 1}),
            (Topic(re.compile(r".*/arm"), Pose), {2}),
            (Topic(re.compile(r".*/arm")), {2}),
            (Topic(re.compile(r".*/arm#geometry.*")), {2}),
        ],
    ),
    Case(
        name="lcm_bytes",
        pubsub_context=lcm_bytes_context,
        topic_values=[
            (Topic("/sensor/temp"), b"temp"),
            (Topic("/sensor/humidity"), b"humidity"),
            (Topic("/robot/arm"), b"arm"),
        ],
        tags={"all", "glob", "regex"},
        glob_patterns=[
            (Topic(topic=Glob("/sensor/*")), {0, 1}),
            (Topic(topic=Glob("/**/arm")), {2}),
            (Topic(topic=Glob("/**")), {0, 1, 2}),
        ],
        regex_patterns=[
            (Topic(re.compile(r"/sensor/.*")), {0, 1}),
            (Topic(re.compile(r".*/arm")), {2}),
        ],
    ),
]

# Build filtered lists for parametrize
all_cases = [c for c in testcases if "all" in c.tags]
glob_cases = [c for c in testcases if "glob" in c.tags]
regex_cases = [c for c in testcases if "regex" in c.tags]


def _topic_matches_prefix(topic: Any, prefix: str = "/") -> bool:
    """Check if topic string starts with prefix.

    LCM uses UDP multicast, so messages from other tests running in parallel
    can leak into subscribe_all callbacks. We filter to only our test topics.
    """
    topic_str = str(topic.topic if hasattr(topic, "topic") else topic)
    return topic_str.startswith(prefix)


@pytest.mark.parametrize("tc", all_cases, ids=lambda c: c.name)
def test_subscribe_all_receives_all_topics(tc: Case[Any, Any]) -> None:
    """Test that subscribe_all receives messages from all topics."""
    received: list[tuple[Any, Any]] = []

    with tc.pubsub_context() as (pub, sub):
        # Filter to only our test topics (LCM multicast can leak from parallel tests)
        sub.subscribe_all(lambda msg, topic: received.append((msg, topic)))
        time.sleep(0.01)  # Allow subscription to be ready

        for topic, value in tc.topic_values:
            pub.publish(topic, value)

        time.sleep(0.01)

        assert len(received) == len(tc.topic_values)

        # Verify all messages were received
        received_msgs = [r[0] for r in received]
        expected_msgs = [v for _, v in tc.topic_values]
        for expected in expected_msgs:
            assert expected in received_msgs


@pytest.mark.parametrize("tc", all_cases, ids=lambda c: c.name)
def test_subscribe_all_unsubscribe(tc: Case[Any, Any]) -> None:
    """Test that unsubscribe stops receiving messages."""
    received: list[tuple[Any, Any]] = []
    topic, value = tc.topic_values[0]

    with tc.pubsub_context() as (pub, sub):
        unsub = sub.subscribe_all(lambda msg, topic: received.append((msg, topic)))
        time.sleep(0.01)  # Allow subscription to be ready

        pub.publish(topic, value)
        time.sleep(0.01)
        assert len(received) == 1

        unsub()

        pub.publish(topic, value)
        time.sleep(0.01)
        assert len(received) == 1  # No new messages


@pytest.mark.parametrize("tc", all_cases, ids=lambda c: c.name)
def test_subscribe_all_with_regular_subscribe(tc: Case[Any, Any]) -> None:
    """Test that subscribe_all coexists with regular subscriptions."""
    all_received: list[tuple[Any, Any]] = []
    specific_received: list[tuple[Any, Any]] = []
    topic1, value1 = tc.topic_values[0]
    topic2, value2 = tc.topic_values[1]

    with tc.pubsub_context() as (pub, sub):
        sub.subscribe_all(
            lambda msg, topic: all_received.append((msg, topic))
            if _topic_matches_prefix(topic)
            else None
        )
        sub.subscribe(topic1, lambda msg, topic: specific_received.append((msg, topic)))
        time.sleep(0.01)  # Allow subscriptions to be ready

        pub.publish(topic1, value1)
        pub.publish(topic2, value2)
        time.sleep(0.01)

        # subscribe_all gets both
        assert len(all_received) == 2

        # specific subscription gets only topic1
        assert len(specific_received) == 1
        assert specific_received[0][0] == value1


@pytest.mark.parametrize("tc", glob_cases, ids=lambda c: c.name)
def test_subscribe_glob(tc: Case[Any, Any]) -> None:
    """Test that glob pattern subscriptions receive only matching topics."""
    for pattern_topic, expected_indices in tc.glob_patterns:
        received: list[tuple[Any, Any]] = []

        with tc.pubsub_context() as (pub, sub):
            sub.subscribe(pattern_topic, lambda msg, topic, r=received: r.append((msg, topic)))
            time.sleep(0.01)  # Allow subscription to be ready

            for topic, value in tc.topic_values:
                pub.publish(topic, value)

            time.sleep(0.01)

            assert len(received) == len(expected_indices), (
                f"Expected {len(expected_indices)} messages for pattern {pattern_topic}, "
                f"got {len(received)}"
            )

            # Verify we received the expected messages
            expected_msgs = [tc.topic_values[i][1] for i in expected_indices]
            received_msgs = [r[0] for r in received]
            for expected in expected_msgs:
                assert expected in received_msgs


@pytest.mark.parametrize("tc", regex_cases, ids=lambda c: c.name)
def test_subscribe_regex(tc: Case[Any, Any]) -> None:
    """Test that regex pattern subscriptions receive only matching topics."""
    for pattern_topic, expected_indices in tc.regex_patterns:
        received: list[tuple[Any, Any]] = []

        with tc.pubsub_context() as (pub, sub):
            sub.subscribe(pattern_topic, lambda msg, topic, r=received: r.append((msg, topic)))

            time.sleep(0.01)

            for topic, value in tc.topic_values:
                pub.publish(topic, value)

            time.sleep(0.01)

            assert len(received) == len(expected_indices), (
                f"Expected {len(expected_indices)} messages for pattern {pattern_topic}, "
                f"got {len(received)}"
            )

            # Verify we received the expected messages
            expected_msgs = [tc.topic_values[i][1] for i in expected_indices]
            received_msgs = [r[0] for r in received]
            for expected in expected_msgs:
                assert expected in received_msgs



# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for pattern matching utilities."""

import re

from dimos.protocol.pubsub.patterns import Glob, pattern_matches


class TestPatternMatchesString:
    """Tests for exact string matching."""

    def test_exact_match(self) -> None:
        assert pattern_matches("/sensor/temp", "/sensor/temp") is True

    def test_no_match(self) -> None:
        assert pattern_matches("/sensor/temp", "/sensor/humidity") is False

    def test_empty_string(self) -> None:
        assert pattern_matches("", "") is True
        assert pattern_matches("", "/sensor") is False

    def test_partial_match_fails(self) -> None:
        assert pattern_matches("/sensor", "/sensor/temp") is False
        assert pattern_matches("/sensor/temp", "/sensor") is False


class TestPatternMatchesGlob:
    """Tests for Glob pattern matching."""

    def test_single_wildcard(self) -> None:
        glob = Glob("/sensor/*")
        assert pattern_matches(glob, "/sensor/temp") is True
        assert pattern_matches(glob, "/sensor/humidity") is True
        assert pattern_matches(glob, "/sensor/") is True

    def test_single_wildcard_no_slash(self) -> None:
        glob = Glob("/sensor/*")
        assert pattern_matches(glob, "/sensor/nested/path") is False

    def test_double_wildcard(self) -> None:
        glob = Glob("/robot/**")
        assert pattern_matches(glob, "/robot/arm") is True
        assert pattern_matches(glob, "/robot/arm/joint1") is True
        assert pattern_matches(glob, "/robot/leg/motor/encoder") is True

    def test_question_mark(self) -> None:
        glob = Glob("/sensor/?")
        assert pattern_matches(glob, "/sensor/a") is True
        assert pattern_matches(glob, "/sensor/1") is True
        assert pattern_matches(glob, "/sensor/ab") is False

    def test_mixed_patterns(self) -> None:
        glob = Glob("/robot/*/joint?")
        assert pattern_matches(glob, "/robot/arm/joint1") is True
        assert pattern_matches(glob, "/robot/leg/joint2") is True
        assert pattern_matches(glob, "/robot/arm/joint12") is False
        assert pattern_matches(glob, "/robot/arm/nested/joint1") is False

    def test_no_wildcards(self) -> None:
        glob = Glob("/exact/path")
        assert pattern_matches(glob, "/exact/path") is True
        assert pattern_matches(glob, "/exact/other") is False

    def test_double_wildcard_middle(self) -> None:
        glob = Glob("/start/**/end")
        # Note: ** becomes .* so /start/**/end requires a / before end
        assert pattern_matches(glob, "/start//end") is True
        assert pattern_matches(glob, "/start/middle/end") is True
        assert pattern_matches(glob, "/start/a/b/c/end") is True


class TestPatternMatchesRegex:
    """Tests for compiled regex pattern matching."""

    def test_simple_regex(self) -> None:
        pattern = re.compile(r"/sensor/\w+")
        assert pattern_matches(pattern, "/sensor/temp") is True
        assert pattern_matches(pattern, "/sensor/123") is True

    def test_regex_anchored(self) -> None:
        pattern = re.compile(r"/sensor/temp")
        assert pattern_matches(pattern, "/sensor/temp") is True
        assert pattern_matches(pattern, "/sensor/temperature") is False

    def test_regex_groups(self) -> None:
        pattern = re.compile(r"/robot/(arm|leg)/joint(\d+)")
        assert pattern_matches(pattern, "/robot/arm/joint1") is True
        assert pattern_matches(pattern, "/robot/leg/joint42") is True
        assert pattern_matches(pattern, "/robot/head/joint1") is False

    def test_regex_optional(self) -> None:
        pattern = re.compile(r"/sensor/temp/?")
        assert pattern_matches(pattern, "/sensor/temp") is True
        assert pattern_matches(pattern, "/sensor/temp/") is True


class TestGlobClass:
    """Tests for the Glob class itself."""

    def test_pattern_property(self) -> None:
        glob = Glob("/sensor/*")
        assert glob.pattern == "/sensor/[^/]*"

    def test_glob_property(self) -> None:
        glob = Glob("/sensor/*")
        assert glob.glob == "/sensor/*"

    def test_repr(self) -> None:
        glob = Glob("/sensor/*")
        assert repr(glob) == "Glob('/sensor/*')"

    def test_double_star_regex(self) -> None:
        glob = Glob("/robot/**")
        assert glob.pattern == "/robot/.*"

    def test_special_chars_escaped(self) -> None:
        glob = Glob("/path.with.dots")
        assert pattern_matches(glob, "/path.with.dots") is True
        assert pattern_matches(glob, "/pathXwithXdots") is False



# Copyright 2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from langchain_core.messages import HumanMessage
import pytest

from dimos.agents.annotation import skill
from dimos.core.module import Module
from dimos.msgs.sensor_msgs import Image
from dimos.utils.data import get_data


class Adder(Module):
    @skill
    def add(self, x: int, y: int) -> str:
        """adds x and y."""
        return str(x + y)


@pytest.mark.integration
@pytest.mark.parametrize("dask", [False, True])
def test_can_call_tool(dask, agent_setup):
    history = agent_setup(
        blueprints=[Adder.blueprint()],
        messages=[HumanMessage("What is 33333 + 100? Use the tool.")],
        dask=dask,
    )

    assert "33433" in history[-1].content


class UserRegistration(Module):
    def __init__(self):
        super().__init__()
        self._first_call = True
        self._use_upper = False

    @skill
    def register_user(self, name: str) -> str:
        """registers a user by name."""

        # If the agent calls with "paul" or "Paul", always say it's the wrong way
        # to force it to try again.

        if self._first_call:
            self._first_call = False
            self._use_upper = not name[0].isupper()

        if self._use_upper and not name[0].isupper():
            return ValueError("Names must start with an uppercase letter.")
        if not self._use_upper and name[0].isupper():
            return ValueError("The names must only use lowercase letters.")

        global _correct_name_registered
        _correct_name_registered = True
        return "User name registered successfully."


@pytest.mark.integration
@pytest.mark.parametrize("dask", [False, True])
def test_can_call_again_on_error(dask, agent_setup):
    history = agent_setup(
        blueprints=[UserRegistration.blueprint()],
        messages=[
            HumanMessage(
                "Register a user named 'Paul'. If there are errors, just try again until you succeed."
            )
        ],
        dask=dask,
    )

    assert any(message.content == "User name registered successfully." for message in history)


class MultipleTools(Module):
    def __init__(self):
        super().__init__()
        self._people = {"Ben": "office", "Bob": "garage"}

    @skill
    def register_person(self, name: str) -> str:
        """Registers a person by name."""
        if name.lower() == "john":
            self._people[name] = "kitchen"
        elif name.lower() == "jane":
            self._people[name] = "living room"
        return f"'{name}' has been registered."

    @skill
    def locate_person(self, name: str) -> str:
        """Locates a person by name."""
        if name not in self._people:
            known_people = list(self._people.keys())
            return (
                f"Error: '{name}' is not registered. People cannot be located until they've "
                f"been registered in the system. People known so far: {', '.join(known_people)}. "
                "Use register_person to register a person."
            )
        return f"'{name}' is located at '{self._people[name]}'."


class NavigationSkill(Module):
    @skill
    def go_to_location(self, description: str) -> str:
        """Go to a location by a description."""
        if description.strip().lower() not in ["kitchen", "living room"]:
            return f"Error: Unknown location description: '{description}'."
        return f"Going to the {description}."


@pytest.mark.integration
def test_multiple_tool_calls_with_multiple_messages(agent_setup):
    history = agent_setup(
        blueprints=[MultipleTools.blueprint(), NavigationSkill.blueprint()],
        messages=[
            HumanMessage(
                "You are a robot assistant. Move to the location where John is. Don't ask me for feedback, just go there."
            ),
            HumanMessage("Nice job. You did it. Now go to the location where Jane is."),
        ],
    )

    # Collect all go_to_location calls from the history
    go_to_location_calls = []
    for message in history:
        if hasattr(message, "tool_calls"):
            for tool_call in message.tool_calls:
                if tool_call["name"] == "go_to_location":
                    go_to_location_calls.append(tool_call)

    # Find the index of the second HumanMessage to split first/second prompt
    second_human_idx = None
    human_count = 0
    for i, message in enumerate(history):
        if isinstance(message, HumanMessage):
            human_count += 1
            if human_count == 2:
                second_human_idx = i
                break

    # Collect go_to_location calls before and after the second prompt
    calls_after_first_prompt = []
    calls_after_second_prompt = []
    for i, message in enumerate(history):
        if hasattr(message, "tool_calls"):
            for tool_call in message.tool_calls:
                if tool_call["name"] == "go_to_location":
                    if i < second_human_idx:
                        calls_after_first_prompt.append(tool_call)
                    else:
                        calls_after_second_prompt.append(tool_call)

    # After the first prompt, go_to_location should be called with "kitchen"
    assert len(calls_after_first_prompt) == 1
    assert "kitchen" in calls_after_first_prompt[0]["args"]["description"].lower()

    # After the second prompt, go_to_location should be called with "living room"
    assert len(calls_after_second_prompt) == 1
    assert "living room" in calls_after_second_prompt[0]["args"]["description"].lower()

    # There should be exactly two go_to_location calls total
    assert len(go_to_location_calls) == 2


@pytest.mark.integration
def test_prompt(agent_setup):
    history = agent_setup(
        blueprints=[],
        messages=[HumanMessage("What is your name?")],
        system_prompt="You are a helpful assistant named Johnny.",
    )

    assert "Johnny" in history[-1].content


class Visualizer(Module):
    @skill
    def take_a_picture(self) -> Image:
        """Takes a picture."""
        return Image.from_file(get_data("cafe-smol.jpg")).to_rgb()


@pytest.mark.integration
def test_image(agent_setup):
    history = agent_setup(
        blueprints=[Visualizer.blueprint()],
        messages=[
            HumanMessage(
                "What do you see? Take a picture using your camera and describe it. "
                "Please mention one of the words which best match the image: "
                "'stadium', 'cafe', 'battleship'."
            )
        ],
        system_prompt="You are a helpful assistant that can use a camera to take pictures.",
    )

    response = history[-1].content.lower()
    assert "cafe" in response
    assert "stadium" not in response
    assert "battleship" not in response



#!/usr/bin/env python3
# Copyright 2025-2026 Dimensional Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Compatibility package for legacy dimos.robot.unitree_webrtc imports."""

from importlib import import_module
import sys

_ALIAS_MODULES = {
    "demo_error_on_name_conflicts": "dimos.robot.unitree.demo_error_on_name_conflicts",
    "depth_module": "dimos.robot.unitree.depth_module",
    "keyboard_teleop": "dimos.robot.unitree.keyboard_teleop",
    "mujoco_connection": "dimos.robot.unitree.mujoco_connection",
    "type": "dimos.robot.unitree.type",
    "unitree_g1_skill_container": "dimos.robot.unitree.g1.skill_container",
    "unitree_skill_container": "dimos.robot.unitree.unitree_skill_container",
}

for alias, target in _ALIAS_MODULES.items():
    sys.modules[f"{__name__}.{alias}"] = import_module(target)
